#!/usr/bin/python3

#
# Copyright (c) 2005 Novell, Inc.
# All Rights Reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General Public
# License as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, contact Novell, Inc.
#
# To contact Novell about this file by physical or electronic mail,
# you may find current contact information at www.novell.com
#
#  Author:  Novell, Inc.
#  Workfile:  dsmlv2.py
#  Contents:  DSMLv2 implementation for ncs and bcc
#

import base64
import clstrlibss
import getopt
import ldap
import re
import sys
import syslog
import string
import traceback
import xml.dom.minidom

########################################################################################
## Globals
########################################################################################

# dsmlv2 namespace
g_DSMLv2Namespace = "urn:oasis:names:tc:DSML:2:0:core"

## global flags used by batch request to determine execution behavior
## Flags for response order
R_SEQUENTIAL = 0
R_UNORDERED  = 1
# Flags for processing order
P_SEQUENTIAL = 0
P_PARALLEL   = 1
# Flags for behavior on error
E_EXIT       = 0
E_RESUME     = 1

# variables for available startup optionts
g_Verbose       = None
g_InputFile     = None
g_OutputFile    = None
g_FormattedName = None
g_LdapUrl       = None
g_Username      = None
g_Password      = None
g_Request       = None

# type flags for search request filter elements and compare assertion
T_BASE             = 0x00
T_AND              = 0x01
T_OR               = 0x02
T_NOT              = 0x03
T_ASSERTION        = 0x04
T_EQUALITY_MATCH   = 0x05
T_SUBSTRINGS       = 0x06
T_GREATER_OR_EQUAL = 0x07
T_LESS_OR_EQUAL    = 0x08
T_PRESENT          = 0x09
T_APPROX_MATCH     = 0x0a
T_EXTENSIBLE_MATCH = 0x0b


# dsml filter tags with corresponding type flag
g_FilterTypeDictionary = {
    "filter"          : T_BASE,
    "and"             : T_AND,
    "or"              : T_OR,
    "not"             : T_NOT,
    "equalityMatch"   : T_EQUALITY_MATCH,
    "substrings"      : T_SUBSTRINGS,
    "greaterOrEqual"  : T_GREATER_OR_EQUAL,
    "lessOrEqual"     : T_LESS_OR_EQUAL,
    "present"         : T_PRESENT,
    "approxMatch"     : T_APPROX_MATCH,
    "extensibleMatch" : T_EXTENSIBLE_MATCH
}

# special characters to escape in response dsml
g_specialChars = (
    ('&','&amp;'),
    ('<','&lt;'),
    ('>','&gt;'),
    ('"','&quot;'),
    ("'",'&apos;')
)

# special characters that must be escaped in a search request filter
g_specialFilterChars = (
    ('\\','\\5c'),
    ('*','\\2a'),
    ('(','\\28'),
    (')','\\29'),
    ('\0','\\00')
)

# eDir object attributes that we know have to be base64 encoded in dsml response
g_Base64Attrs = {
    'networkAddress' : None,
    'nCSNetworkAddress' : None,
    'nCSCRMPreferredNodes' : None,
    'GUID' : None,
    'objectGUID' : None
}

# table to map ldap execptions to corresponding dsml description (LDAPResultCode) and value
g_LDAPErrorTable = {
    'ADMINLIMIT_EXCEEDED'            : ["adminLimitExceeded",          0x0b],
    'AFFECTS_MULTIPLE_DSAS'          : ["affectMultipleDSAs",          0x47],
    'ALIAS_DEREF_PROBLEM'            : ["aliasDereferencingProblem",   0x24],
    'ALIAS_PROBLEM'                  : ["aliasProblem",                0x21],
    'ALREADY_EXISTS'                 : ["entryAlreadyExists",          0x44],
    'AUTH_UNKNOWN'                   : ["authMethodNotSupported",      0x07],
    'BUSY'                           : ["busy",                        0x33],
    'CLIENT_LOOP'                    : ["other",                       0x50],
    'COMPARE_FALSE'                  : ["compareFalse",                0x05],
    'COMPARE_TRUE'                   : ["compareTrue",                 0x06],
    'CONFIDENTIALITY_REQUIRED'       : ["confidentialityRequired",     0x0d],
    'CONNECT_ERROR'                  : ["couldNotConnect",             0x50],
    'CONSTRAINT_VIOLATION'           : ["constraintViolation",         0x13],
    'CONTROL_NOT_FOUND'              : ["other",                       0x50],
    'DECODING_ERROR'                 : ["other",                       0x50],
    'ENCODING_ERROR'                 : ["other",                       0x50],
    'FILTER_ERROR'                   : ["other",                       0x50],
    'INAPPROPRIATE_AUTH'             : ["inappropriateAuthentication", 0x30],
    'INAPPROPRIATE_MATCHING'         : ["inappropriateMatching",       0x12],
    'INSUFFICIENT_ACCESS'            : ["insufficientAccessRights",    0x32],
    'INVALID_CREDENTIALS'            : ["invalidCredentials",          0x31],
    'INVALID_DN_SYNTAX'              : ["invalidDNSyntax",             0x22],
    'INVALID_SYNTAX'                 : ["invalidAttributeSyntax",      0x15],
    'IS_LEAF'                        : ["other",                       0x23],
    'LOCAL_ERROR'                    : ["other",                       0x50],
    'LOOP_DETECT'                    : ["loopDetect",                  0x36],
    'MORE_RESULTS_TO_RETURN'         : ["other",                       0x50],
    'NAMING_VIOLATION'               : ["namingViolation",             0x40],
    'NO_OBJECT_CLASS_MODS'           : ["other",                       0x50],
    'NOT_ALLOWED_ON_NONLEAF'         : ["notAllowedOnNonLeaf",         0x42],
    'NOT_ALLOWED_ON_RDN'             : ["notAllowedOnRDN",             0x43],
    'NOT_SUPPORTED'                  : ["other",                       0x50],
    'NO_MEMORY'                      : ["other",                       0x50],
    'NO_OBJECT_CLASS_MODS'           : ["objectClassModsProhibited",   0x45],
    'NO_RESULTS_RETURNED'            : ["other",                       0x50],
    'NO_SUCH_ATTRIBUTE'              : ["noSuchAttribute",             0x10],
    'NO_SUCH_OBJECT'                 : ["noSuchObject",                0x20],
    'OBJECT_CLASS_VIOLATION'         : ["objectClassViolation",        0x41],
    'OPERATIONS_ERROR'               : ["other",                       0x01],
    'OTHER'                          : ["other",                       0x50],
    'PARAM_ERROR'                    : ["other",                       0x50],
    'PARTIAL_RESULTS'                : ["other",                       0x50],
    'PROTOCOL_ERROR'                 : ["protocolError",               0x02],
    'RESULTS_TOO_LARGE'              : ["other",                       0x46],
    'SASL_BIND_IN_PROGRESS'          : ["saslBindInProgress",          0x0e],
    'SERVER_DOWN'                    : ["couldNotConnect",             0x50],
    'SIZELIMIT_EXCEEDED'             : ["sizeLimitExceeded",           0x04],
    'STRONG_AUTH_NOT_SUPPORTED'      : ["other",                       0x50],
    'STRONG_AUTH_REQUIRED'           : ["strongAuthRequired",          0x08],
    'TIMELIMIT_EXCEEDED'             : ["timeLimitExceeded",           0x03],
    'TIMEOUT'                        : ["other",                       0x50],
    'TYPE_OR_VALUE_EXISTS'           : ["attributeOrValueExists",      0x14],
    'UNAVAILABLE'                    : ["unavailable",                 0x34],
    'UNAVAILABLE_CRITICAL_EXTENSION' : ["unavailableCriticalExtension",0x0c],
    'UNDEFINED_TYPE'                 : ["undefinedAttributeType",      0x11],
    'UNWILLING_TO_PERFORM'           : ["unwillingToPerform",          0x35],
    'USER_CANCELLED'                 : ["other",                       0x50]
}

########################################################################################
## Functions
########################################################################################

def attributeDescriptionValue(s):
    """
    Make sure that a string follows correct pattern for dsml AttributeDescriptionValue type

    Throws an exception if the string 's' does not match the pattern
    """
    if re.compile("""^((([0-2](\.[0-9]+)+)|([a-zA-Z]+([a-zA-Z0-9]|[-])*))(;([a-zA-Z0-9]|[-])+)*)$""").match(s):
        return s
    else:
        raise Exception ("Attribute value, '"+s+"' does not have correct format.")

def base64Encode(s):
    """
    Return s (string) base64 encoded
    """
    return base64.encodestring(s)

def buildLDAPResult(requestRef,tag):
    """
    Generic function to return an dsml LDAPResult response (string) for a given dsml request (Request)

    Parameters:
    requestRef
        reference to an instance of Request class, or derivative thereof
    tag
        name of dsml response for which the response is built

    Return format:
        <tag requestID=... code=... descr=...><errorMessage>...</errorMessage></tag>
    """
    if requestRef._attemptedService:
        s = "<"+tag
        if requestRef._requestID:
            s += " requestID=\""+escapeChars(str(requestRef._requestID))+"\""
        s += "><resultCode code=\""+escapeChars(str(requestRef._code))+"\" descr=\""+escapeChars(requestRef._descr)+"\"/>"
        if requestRef._errorMessage:
            s += "<errorMessage>"+escapeChars(requestRef._errorMessage)+"</errorMessage>"
        s += "</"+tag+">"
        return s
    else:
        s = "<errorResponse type=\"notAttempted\""
        if requestRef._requestID:
            s += " requestID=\""+escapeChars(str(requestRef._requestID))+"\""
        s += "/>"
        return s

def debugLog(s):
    """
    Print a message to both stderr and syslog when invoked with g_Verbose set (-v switch)
    """
    if g_Verbose:
        lines = s.split('\n')
        for line in lines:
            syslog.syslog(syslog.LOG_DEBUG,line)
        sys.stderr.write(s)

def elementToXML(element):
    """
    Returns string representation of dom structure from element with white-space stripped
    from between tags
    """
    return re.sub(r'>\s+<','><',element.toxml())

def escapeChars(s):
    """
    Return s (string) with special characters escaped

    Used to escape special characters from strings that will be encapsulated within dsml responses
    """
    for char,entity in g_specialChars:
        s = string.replace(s,char,entity)
    return s

def escapeFilterChars(s):
    """
    Return s (string) with special filter characters escaped

    Used to escape special characters from ldap search filters
    """
    if (s):
        for char,entity in g_specialFilterChars:
            s = string.replace(s,char,entity)
        return s
    return None

def getElementAttribute(element,attr,require=True):
    """
    Return a string value for an attribute of element whose name is specified by attr

    If require is true, and attr was not found in element, an exception is raise.
    """
    ret = element.getAttributeNode(attr)
    if ret == None:
        if require:
            raise Exception ( "Did not find required attribute '"+attr+"' for element "+elementToXML(element))
        else:
            return None
    return ret.nodeValue

def getElementAttributes(element,attrList):
    """
    Return an array of values for attributes whose names are contained in attrList

    Parameters:
    attrList
        tuple with the format (('<attr name>',<boolean>),...)
        <boolean> is used to specify if the attribute has to be found.  If True,
        and <attr name> was not found, an exception is raise.
    """
    ret = []
    for attr,require in attrList:
        ret.append(getElementAttribute(element,attr,require))
    return ret

def getElementValue(element,required=True):
    """
    Function to return the text section of an element (string).

    Parameters:
    element
        Dom element for which text section is parsed
    required
        Is the text section required

    Checks to see if element has a base64/utf-8 encoding tag, if so,
    to decode the text section.

    If required is True, and the element does not have a text node as first child, an exception is
    raised, otherwise empty string is returned.
    """
    if element.firstChild:
        if element.firstChild.nodeType == xml.dom.minidom.Node.TEXT_NODE:
            encoding = element.getAttribute("encoding")
            if encoding:
                if encoding == "base64":
                    return base64.decode(element.firstChild.data)
                elif encoding == "utf-8":
                    return str(unicode(element.firstChild.data))
                else:
                    raise Exception ( "Don't know how to decode encoding '"+encoding+"' for element: "+elementToXML(element))
            else:
                return str(element.firstChild.data)
        elif element.firstChild.nodeType == xml.dom.minidom.Node.CDATA_SECTION_NODE:
            return str(element.firstChild.data)
        else:
            raise Exception ( "Format of element is incorrect, expecting a text node: "+elementToXML(element))
    elif required:
        raise Exception ( "Element does not have a child node: "+elementToXML(element))
    else:
        return ""

def handleLDAPError(className,args):
    """
    Process an exception that was raised when executing ldap operations

    Parameters:
    className
        class name of exception to handle
    args
        dictionary info of exception - usually from sys.sys_exc()[1]
        in the case of ldap exceptions, possibly has the form { 'desc' : <string>, 'info' : <string>, 'matched' : <string> }
        (possibly because the matched entry might not be present)

    Returns an array of length three with the following format:
    [<dsml-error-descr>,<ldap-error-code>,<descriptive error message, or "" if no message>]
    """
    try:
        ret = [g_LDAPErrorTable[className][0],g_LDAPErrorTable[className][1]]
        s = ""
        for i, j in args.iteritems():
            if s != "":
                s += " :: "
            s += str(j)
        ret.append(s)
    except:
        return (str(-1),"Unknown error.",str(args))
    if len(ret) == 2:
        ret.append("")
    assert len(ret) == 3, "Return array for error if not of length 3.  "+str(ret)
    return ret

def hasNonPrint(s):
    """
    Returns True if string 's' contains a non-print ascii character,
    otherwise, False

    Parameters:
    s
        String to check
    """
    for i in range(len(s)):
        if not isPrint(s[i]):
            return True
    return False

def isPrint(c):
    """
    Returns True is character 'c' is a printable ascii character,
    otherwise, False
    """
    if (31 < ord(c) < 127) | (c == '\t') | (c == '\r') | (c == '\n'):
        return True
    return False

def needsBase64Encoding(attr,value):
    """
    Check to see if value (string) needs base64 encoding before being sent in dsml response

    Parameters:
    attr
        Attribute name to check against known attributes that need to be encoded
    value
        value of attribute

    If value contains any non-print ascii characters, or attr is a key in g_Base64Attrs, True
    is returned, otherwise, False is returned
    """
    if g_Base64Attrs.has_key(value):
        return True
    elif hasNonPrint(value):
        return True
    return False

def needsCDATAWrapping(s):
    """
    Check to see if s (string) has any special xml characters.
    """
    return re.compile("[<>'\"&]").match(s)

def newFilterObject(element):
    """
    Return a new search filter object based upon the given dsml snip
    """
    type = typeByTag(element.localName)
    if (type <= T_NOT):
        return Filter(element)
    elif (type == T_SUBSTRINGS):
        return SubstringFilter(element)
    elif (type == T_PRESENT):
        return AttributeDescription(element)
    elif (type == T_EXTENSIBLE_MATCH):
        return MatchingRuleAssertion(element)
    else:
        return AttributeValueAssertion(element)

def parseDsmlModification(element):
    """
    Parse a <modification> (DsmlModification) element from dsml snip contained within dom node 'element'

    Returns a tuple (modlist) with length three that can be passed directly to ldap.modify()
    """
    s = getElementAttribute(element,'operation',True)
    if s == "add":
        mod_op = ldap.MOD_ADD
    elif s == "delete":
        mod_op = ldap.MOD_DELETE
    elif s == "replace":
        mod_op = ldap.MOD_REPLACE
    else:
        raise Exception ( "Invalid 'operation' value for modify request, '"+operation+"': "+elementToXML(element))

    mod_type = attributeDescriptionValue(getElementAttribute(element,'name',True))

    mod_vals = []
    for value in element.getElementsByTagNameNS(g_DSMLv2Namespace,"value"):
        mod_vals.append(getElementValue(value))
    return (mod_op,mod_type,mod_vals)

def printHelp():
    """
    Prints a help message to sys.stderr
    """
    message = "DSML Version 2 Parser - supports a subset of the DSMLv2 specifications as approved by the OASIS standard April 30, 2002.\n"
    message += "usage: "+str(sys.argv[0])+" [-h] [-v] [[-p <dsml request>]|[-i <input file>]] [-o <output file>] [ -f <formatted-out-file>] [-l <ldap url>] [-u <user fdn>] [-p <password>]\n\n"
    message += "\t -h Print this help message\n"
    message += "\t -v Be verbose. (dumps ldap happenings)\n"
    message += "\t -i Input file from which dsml request is read. (stdin if not specified)\n"
    message += "\t -o Output file to which dsml response is written. (stdout if not specified)\n"
    message += "\t -r Dsml request passed as an arguement.\n"
    message += "\t -f Output file to which pretty formatted dsml response is written. (None is not specified)\n"
    message += "If the three options below are not specified, their values come from the ncs secret store.\n"
    message += "\t -l ldap url to use for binding.  Can be a comma separated list. (ldaps://127.0.0.1:636,...)\n"
    message += "\t -u user fully distinguished name to use in ldap binding. (cn=admin,o=group)\n"
    message += "\t -p password to use in ldap binding.\n\n"
    sys.stderr.write(message)

def processSearchAttributes(attrsEl):
    """
    Process attribute elements of a search request

    Parameters:
    attrsEl
        'attributes' element of search request

    Returns an array of attributes that can be passed to the ldap.search() method.
    """
    ret = []
    attrElList = attrsEl.getElementsByTagNameNS(g_DSMLv2Namespace,"attribute")
    for attrEl in attrElList:
        ret.append(str(attributeDescriptionValue(getElementAttribute(attrEl,"name"))))
    return ret

def typeByTag(tag):
    """
    Return search request filter (int) type for the given tag
    """
    if g_FilterTypeDictionary.has_key(tag):
        return g_FilterTypeDictionary[tag]
    else:
        raise Exception ( "Unknown tag: "+str(tag))

def wrapInCDATA(s):
    """
    Wraps s (string) in an xml CDATA secion
    """
    return "<![CDATA["+s+"]]>"

def writeDsmlResponse(response):
    """
    Write dsml result to appropriate place

    Parameters:
        response
            dom document representing a dsml response
    """
    # encode result in utf-8 and send to output
    if g_OutputFile:
        f = open(g_OutputFile,'w')
        f.write(response.toxml("utf-8"))
        f.close()
    else:
        sys.stdout.write(response.toxml("utf-8"))
    # dump pretty formatted results to file
    if (g_FormattedName):
        f = open(g_FormattedName,'w')
        f.write(response.toprettyxml())
        f.close()
    # send pretty formatted result to stderr for debugging
    debugLog("+#### dsml response ####\n"+response.toprettyxml()+"-#######################\n")

########################################################################################
## Request classes
##
## used to abstract request types from DOM's
## and service the respective LDAP request
########################################################################################
class Request:
    """
    Base class for DSMLv2 Requests.
    """
    def __init__(self,tagName,element):
        self._tagName = tagName
        assert element.localName == self._tagName, "Element indicates wrong request type: "+elementToXML(element)
        self._requestID = None
        self._error = None
        self._code = None
        self._descr = None
        self._errorMessage = None
        self._attemptedService = False

        # get requestID if present
        self._requestID = getElementAttribute(element,"requestID",False)
        if (self._requestID != None) & (self._requestID == "0"):
            raise Exception ( "'"+tagName+"' cannot have a request id of zero(0): "+elementToXML(element))

    def serviceRequest(self,ld):
        """
        Process the request over the ldap connection provided by ld
        """
        self._error = False
        self._code = 0
        self._descr = "success"
        self._attemptedService = True
        return True

    def response(self):
        """
        Build and return dsml response (string) for request
        """
        return buildLDAPResult(self,self._tagName.replace("Request","Response"))

class AddRequest(Request):
    """
    Class for handling add requests
    """
    def __init__(self,element):
        """
        Initialize class

        Parameters:
        element
            An xml.dom.minidom.Element that contains the dsmlv2 structure of a single add request.
        """
        Request.__init__(self,"addRequest",element)
        # get and assert DN
        self._dn = getElementAttribute(element,"dn")
        # process all name/value attribute pairings
        # the dsml spec allows for multiple values for a single named attribute, not sure how handle that as far as passing to ldap
        self._attrs = []
        attrs = element.getElementsByTagNameNS(g_DSMLv2Namespace,"attr")
        for attr in attrs:
            name = attr.getAttribute("name")
            if name == None:
                raise Exception ( "Encounterd a value node with no name.")
            values = []
            vals = attr.getElementsByTagNameNS(g_DSMLv2Namespace,"value")
            for val in vals:
                textNode = val.firstChild
                if textNode == None:
                    continue
                if textNode.nodeType != xml.dom.minidom.Node.TEXT_NODE:
                    raise Exception ( "encountered incorrect node type within attribute, '"+name+"', structure - should be a text node: "+elementToXML(attr))
                values.append(str(textNode.data))
            self._attrs.append((name,values))

    def serviceRequest(self,ld):
        """
        Process the add request over the ldap connection provided by ld
        """
        Request.serviceRequest(self,ld)
        try:
            ld.add_s(self._dn,self._attrs)
        except:
            self._error = True
            self._descr, self._code, self._errorMessage = handleLDAPError(str(sys.exc_info()[0]).split('.')[1],sys.exc_info()[1].args[0])
        return not self._error

class CompareRequest(Request):
    """
    Class for handling compare requests
    """
    def __init__(self,element):
        """
        Initialize class

        Parameters:
        element
            An xml.dom.minidom.Element that contains the dsmlv2 structure of a single compare request.
        """
        Request.__init__(self,"compareRequest",element)
        # get and assert DN
        self._dn = getElementAttribute(element,"dn",True)

        # get assertion information
        assertions = element.getElementsByTagNameNS(g_DSMLv2Namespace,"assertion")
        if len(assertions) != 1:
            raise Exception ( "'compareRequest' must have exactly one 'assertion' element: "+elementToXML(element))
        self._attrName = attributeDescriptionValue(getElementAttribute(assertions[0],"name"))
        self._attrValue = None
        children = assertions[0].childNodes
        for i in range(0,children.length):
            child = children.item(i)
            if child.nodeType == xml.dom.minidom.Node.ELEMENT_NODE:
                if child.localName == "value":
                    if self._attrValue:
                        raise Exception ( "Only one 'value' element allowed: "+elementToXML(child))
                    else:
                        self._attrValue = escapeFilterChars(getElementValue(child,False))
                else:
                    raise Exception ( "Unknown element: "+elementToXML(child))
            else:
                pass
        if self._attrValue == None:
            raise Exception ( "'"+self._tagName+"' must have a 'value' element: "+elementToXML(element))

    def serviceRequest(self,ld):
        """
        Process the compare request over the ldap connection provided by ld
        """
        Request.serviceRequest(self,ld)
        try:
            if ld.compare_s(self._dn,self._attrName,self._attrValue) == 1:
                self._descr, self._code = g_LDAPErrorTable['COMPARE_TRUE'][0],g_LDAPErrorTable['COMPARE_TRUE'][1]
            else:
                self._descr, self._code = g_LDAPErrorTable['COMPARE_FALSE'][0],g_LDAPErrorTable['COMPARE_FALSE'][1]
        except:
            self._error = True
            self._descr, self._code, self._errorMessage = handleLDAPError(str(sys.exc_info()[0]).split('.')[1],sys.exc_info()[1].args[0])
        return not self._error

class DelRequest(Request):
    """
    Class for handling delete requests
    """
    def __init__(self,element):
        """
        Initialize class

        Parameters:
        element
            An xml.dom.minidom.Element that contains the dsmlv2 structure of a single delete request.
        """
        Request.__init__(self,"delRequest",element)
        # get and assert DN
        self._dn = getElementAttribute(element,"dn",True)

    def serviceRequest(self,ld):
        """ Process the addRequest over the ldap connection provided by ld """
        Request.serviceRequest(self,ld)
        try:
            ld.delete_s(self._dn)
        except:
            self._error = True
            self._descr, self._code, self._errorMessage = handleLDAPError(str(sys.exc_info()[0]).split('.')[1],sys.exc_info()[1].args[0])
        return not self._error

class ModDNRequest(Request):
    """
    Class for handling modify DN requests
    """
    def __init__(self,element):
        """
        Initialize class

        Parameters:
        element
            An xml.dom.minidom.Element that contains the dsmlv2 structure of a single modify DN request.
        """
        Request.__init__(self,"modDNRequest",element)
        # Attribute for element
        a = (
            ("dn",True),
            ("newrdn",True),
            ("deleteoldrdn",False),
            ("newSuperior",False)
        )
        # parse request attributes
        self._dn, self._newrdn, self._deleteoldrdn, self._newSuperior = getElementAttributes(element,a)

    def serviceRequest(self,ld):
        """
        Process the modify DN request over the ldap connection provided by ld
        """
        Request.serviceRequest(self,ld)
        delete = 1
        if (self._deleteoldrdn != None) & (self._deleteoldrdn != "true"):
            delete = 0
        try:
            if self._newSuperior:
                raise Exception ( "Not supporting 'newSuperior' attribute for 'modifyDNRequest'.")
            else:
                ld.modrdn_s(self._dn,self._newrdn,delete)
        except:
            self._error = True
            self._descr, self._code, self._errorMessage = handleLDAPError(str(sys.exc_info()[0]).split('.')[1],sys.exc_info()[1].args[0])
        return not self._error

class ModifyRequest(Request):
    """
    Class for handling modify requests
    """
    def __init__(self,element):
        """
        Initialize class

        Parameters:
        element
            An xml.dom.minidom.Element that contains the dsmlv2 structure of a single modify request.
        """
        Request.__init__(self,"modifyRequest",element)
        self._dn = getElementAttribute(element,"dn",True)
        self._modlist = []
        mods = element.getElementsByTagNameNS(g_DSMLv2Namespace,"modification")
        for mod in mods:
            self._modlist.append(parseDsmlModification(mod))

    def serviceRequest(self,ld):
        """
        Process the modify request over the ldap connection provided by ld
        """
        Request.serviceRequest(self,ld)
        try:
            ld.modify_s(self._dn,self._modlist)
        except:
            self._error = True
            self._descr, self._code, self._errorMessage = handleLDAPError(str(sys.exc_info()[0]).split('.')[1],sys.exc_info()[1].args[0])
        return not self._error

class SearchRequest(Request):
    """
    Class for handling search requests
    """
    def __init__(self,element):
        """
        Initialize class

        Parameters:
        element
            An xml.dom.minidom.Element that contains the dsmlv2 structure of a single search request.
        """
        Request.__init__(self,"searchRequest",element)
        self._results = None
        self._filter = "(objectClass=*)"
        self._attributes = None

        # parse dn - required yes
        self._dn = getElementAttribute(element,"dn",True)

        # parse scope - required yes
        scope = getElementAttribute(element,"scope",True)
        if scope == "baseObject":
            self._scope = ldap.SCOPE_BASE
        elif scope == "singleLevel":
            self._scope = ldap.SCOPE_ONELEVEL
        elif scope == "wholeSubtree":
            self._scope = ldap.SCOPE_SUBTREE
        else:
            raise Exception ( "Unknown scope specifier: '"+scope+"' "+elementToXML(element))

        # parse derefAliases - required yes
        derefAliases = getElementAttribute(element,"derefAliases",True)
        if derefAliases == "neverDerefAliases":
            self._derefAliases = ldap.DEREF_NEVER
        elif derefAliases == "derefInSearching":
            self._derefAliases = ldap.DEREF_SEARCHING
        elif derefAliases == "derefFindingBaseObj":
            self._derefAliases = ldap.DEREF_FINDING
        elif derefAliases == "derefAlways":
            self._derefAliases = ldap.DEREF_ALWAYS
        else:
            raise Exception ( "Unknown dereference alias scecifier: '"+derefAliases+"' "+elementToXML(element))

        # parse size limit - required no / default is 0
        sizeLimit = getElementAttribute(element,"sizeLimit",False)
        if sizeLimit == None:
            self._sizeLimit = 0
        else:
            self._sizeLimit = int(sizeLimit)

        # parse time limit - required no / default 0
        timeLimit = getElementAttribute(element,"timeLimit",False)
        if timeLimit == None:
            self._timeLimit = 0
        else:
            self._sizeLimit = int(timeLimit)

        # parse types only - required no / default false
        typesOnly = getElementAttribute(element,"typesOnly",False)
        if typesOnly != None:
            if typesOnly == "false":
                self._typesOnly = 0
            elif typesOnly == "true":
                self._typesOnly = 1
            else:
                raise Exception ( "Unknown search types only specifier: "+typesOnly)
        else:
            self._typesOnly = 0

        # parse filter
        filter = element.getElementsByTagNameNS(g_DSMLv2Namespace,"filter")
        if filter.length > 1:
            raise Exception ( "Search request can have only one 'filter' element: "+elementToXML(element))
        if filter:
            self._filter = newFilterObject(filter.item(0)).value()

        # parse attributes
        attributes = element.getElementsByTagNameNS(g_DSMLv2Namespace,"attributes")
        if (attributes != None) & (attributes.length > 1):
            raise Exception ( "Search request can have zero or one 'attributes' element: "+elementToXML(element))
        if attributes.length == 1:
            self._attributes = processSearchAttributes(attributes.item(0))

    def serviceRequest(self,ld):
        """
        Process the search request over the ldap connection provided by ld
        """
        Request.serviceRequest(self,ld)
        try:
            ld.deref = self._derefAliases
            ld.sizelimit = self._sizeLimit
            ld.timelimit = self._timeLimit
            self._results = ld.search_s(self._dn,self._scope,filterstr=self._filter,attrlist=self._attributes)
            debugLog("return from ldap:\n"+str(self._results))
        except:
            self._error = True
            self._descr, self._code, self._errorMessage = handleLDAPError(str(sys.exc_info()[0]).split('.')[1],sys.exc_info()[1].args[0])
        return not self._error

    def response(self):
        """
        Build and return the dsml response for this search request
        """
        if self._attemptedService:
            s = "<searchResponse"
            if self._requestID:
                s += " requestID=\""+escapeChars(str(self._requestID))+"\""
            s += ">"
            if self._results:
                for result in self._results:
                    s += "<searchResultEntry dn=\""+escapeChars(result[0])+"\">"
                    for attr,values in result[1].iteritems():
                        s += "<attr name=\""+escapeChars(attr)+"\">"
                        for value in values:
                            if needsBase64Encoding(attr,value):
                                s += "<value encoding=\"base64\">"+base64Encode(value)+"</value>"
                            elif needsCDATAWrapping(value):
                                s += "<value>"+wrapInCDATA(value)+"</value>"
                            else:
                                s += "<value>"+value+"</value>"
                        s += "</attr>"
                    s += "</searchResultEntry>"
            s += "<searchResultDone>"
            s += "<resultCode code=\""+escapeChars(str(self._code))+"\" descr=\""+escapeChars(self._descr)+"\"/>"
            if self._errorMessage:
                s += "<errorMessage>"+escapeChars(self._errorMessage)+"</errorMessage>"
            s += "</searchResultDone>"
            s += "</searchResponse>"
            return s
        else:
            s = "<errorResponse type=\"notAttempted\""
            if self._requestID:
                s += " requestID=\""+escapeChars(str(self._requestID))+"\""
            s += "/>"
            return s

class BatchRequest(Request):
    """
    Class for handling batch requests.
    """
    def __init__(self,element):
        """
        Create a batch request structure from the dom node 'element'
        """
        self._requests = []

        # Set the default attributes for the batchRequest
        self._responseOrder = R_SEQUENTIAL
        self._processing = R_SEQUENTIAL
        self._onError = E_EXIT
        self._validRequest = True

        # Process this batchRequest structure, any exceptions that are caught "should be" syntax errors
        # that are reported back to the client as a malformedRequest
        try:
            # Parse attributes for batchRequest
            Request.__init__(self,"batchRequest",element)
            assert element.localName == "batchRequest", "Trying to instantiate BatchRequest from non-batchRequest element."
            assert element.namespaceURI == g_DSMLv2Namespace, "Invalid namespace for batchRequest"
            for i in range(0,element.attributes.length):
                attr = element.attributes.item(i)
                if attr.localName == "responseOrder":
                    if attr.nodeValue == "sequential":
                        self._responseOrder = R_SEQUENTIAL
                    elif attr.nodeValue == "unordered":
                        self._responseOrder = R_UNORDERED
                    else:
                        raise Exception ( "Unknown responseOrder value: "+attr.nodeValue)
                elif attr.localName == "processing":
                    if attr.nodeValue == "sequential":
                        self._processing = P_SEQUENTIAL
                    elif attr.nodeValue == "parallel":
                        self._processing = P_PARALLEL
                    else:
                        raise Exception ( "Unknown processing value: "+attr.nodeValue)
                elif attr.localName == "onError":
                    if attr.nodeValue == "exit":
                        self._onError = E_EXIT
                    elif attr.nodeValue == "resume":
                        self._onError = E_RESUME
                    else:
                        raise Exception (  "Unknown onError value: "+attr.nodeValue)
                elif (attr.localName == "xsi") | (attr.localName == "xsd") | (attr.localName == "dsml") | (attr.localName == "xmlns") | (attr.localName == "requestID"):
                    pass
                else:
                    raise Exception ( "Unkown attribute: "+str(attr.localName))

            # Parse child elements of batchRequest
            for child in element.childNodes:
                if child.nodeType == xml.dom.minidom.Node.ELEMENT_NODE:
                    if child.namespaceURI != g_DSMLv2Namespace:
                        raise Exception ( "Invalid namespace, "+str(child.namespaceURI)+", for element: "+str(child.localName))
                    if child.localName == "addRequest":
                        self._requests.append(AddRequest(child))
                    elif child.localName == "abanonRequest":
                        raise Exception ( "Request type, abandonRequest, not supported.")
                    elif child.localName == "compareRequest":
                        self._requests.append(CompareRequest(child))
                    elif child.localName == "delRequest":
                        self._requests.append(DelRequest(child))
                    elif child.localName == "extendedRequest":
                        raise Exception ( "Request type, extendedRequest, not supported.")
                    elif child.localName == "modDNRequest":
                        self._requests.append(ModDNRequest(child))
                    elif child.localName == "modifyRequest":
                        self._requests.append(ModifyRequest(child))
                    elif child.localName == "searchRequest":
                        self._requests.append(SearchRequest(child))
                    elif child.localName == "authRequest":
                        raise Exception ( "proxy authentication not supported")
                    else:
                        raise Exception ( "Parsing error: Unknown element: "+str(child.localName))
                elif (child.nodeType == xml.dom.minidom.Node.COMMENT_NODE)| (child.nodeType == xml.dom.minidom.Node.TEXT_NODE):
                    pass
                else:
                    raise Exception ( "Parsing error: unexpected node found: "+str(child.localName))
        except Exception as e:
            self._validRequest = False
            self._errorMessage = ""
            for i in e.args:
                self._errorMessage += i

    def serviceRequest(self,ldapUrl,ldapUser,ldapPassword):
        """
        Process all requests with the given ldap url, user name and password
        """
        assert self._validRequest
        assert self._requests
        ld = None
        try:
            ld = ldap.initialize(g_LdapUrl)
            ld.simple_bind(g_Username,g_Password)
        except:
            self._error = True
            self._descr, self._code, self._errorMessage = handleLDAPError(str(sys.exc_info()[0]).split('.')[1],sys.exc_info()[1].args[0])
            return False

        # treating all requests as though they are sequential
#       if self._responseOrder == R_SEQUENTIAL:
        for request in self._requests:
            if (not request.serviceRequest(ld)) & (self._onError == E_EXIT):
                break

        if ld:
            try:
                ld.unbind()
            except:
                pass

    def response(self):
        """
        Build and return the dsml response for this batch request
        """
        s = "<batchResponse"
        if self._requestID != None:
            s += " requestID=\""+str(self._requestID)+"\""
        s += " xmlns=\""+g_DSMLv2Namespace+"\">"
        if (not self._error):
            if (self._validRequest):
                for i in self._requests:
                    s += i.response()
            else:
                s += "<errorResponse type=\"malformedRequest\"><errorMessage>"+escapeChars(re.sub(r'>\s+','><',self._errorMessage))+"</errorMessage></errorResponse>"
        else:
            s += "<errorResponse type=\""+self._descr+"\"><errorMessage>"+escapeChars(re.sub(r'>\s+','><',self._errorMessage))+"</errorMessage></errorResponse>"
        s += "</batchResponse>"
        return s


########################################################################################
## Filter classes used in processing search requests
########################################################################################
class Filter:
    """
    Used for filter,and,or,not tags

    Parameters:
        element
            Has the form:
                <filter>...</filter>
            Or:
                <and>...</and>
            Or:
                <or>...</or>
            Or:
                <not>...</not>
    """
    def __init__(self,element):
        self._children = []
        self._ldapFilter = ""
        self._type = typeByTag(element.localName)
        assert self._type <= T_NOT, "Should not have instantiated a filter object for type '"+element.localName+"'"
        self._tagName = element.localName
        children = element.childNodes
        for child in children:
            if child.nodeType == xml.dom.minidom.Node.ELEMENT_NODE:
                if self._type == T_NOT:
                    assert len(self._children) == 0, "'not' filters can have only one nested filter: cannot add '"+child.localName+"' to filter chain."
                    self._children.append(newFilterObject(child))
                else:
                    self._children.append(newFilterObject(child))

        self._ldapFilter = ""
        if self._type > T_BASE:
            self._ldapFilter += "("
        if self._type == T_AND:
            self._ldapFilter += "&"
        elif self._type == T_OR:
            self._ldapFilter += "|"
        elif self._type == T_NOT:
            self._ldapFilter += "!"
        for child in self._children:
            self._ldapFilter += child.value()
        if self._type > T_BASE:
            self._ldapFilter += ")"

    def value(self):
        return self._ldapFilter

class AttributeDescription(Filter):
    """
    Created for present tag

    Parameters:
        element
            Has the form:
                <present name="cn"/>
            Which, in ldap syntaax, translates to:
                (cn=*)
    """
    def __init__(self,element):
        self._type = typeByTag(element.localName)
        self._tagName = element.localName
        self._name = attributeDescriptionValue(getElementAttribute(element,"name"))
        children = element.childNodes
        for child in children:
            if child.nodeType == xml.dom.minidom.Node.ELEMENT_NODE:
                raise Exception ( "'"+self._tagName+"' should not have any child elements: "+elementToXML(element))

        self._ldapFilter = "("+self._name+"=*)"

class AttributeValueAssertion(Filter):
    """
    Created for equalityMatch,greaterOrEqual,lessOrEqual,approxMatch tags

    Paramters:
        element
            Has the form:
                <equalityMatch name="cn">
                    <value>Novell</value>
                </equalityMatch>
            Which, in ldap syntax, translates to:
                (cn=Novell)

            Or has the form:
                <greaterOrEqual name="cn">
                    <value>Novell</value>
                </greaterOrEqual>
            Which, in ldap syntax, translates to:
                (cn>=Novell)

            Or has the form:
                <lessOrEqual name="cn">
                    <value>Novell</value>
                </lessOrEqual>
            Which, in ldap syntax, translates to:
                (cn<=Novell)

            Or has the form:
                <approxMatch name="cn">
                    <value>Novell</value>
                </approxMatch>
            Which, in ldap syntax, translates to:
                (cn~=Novell)
    """
    def __init__(self,element):
        self._type = typeByTag(element.localName)
        self._tagName = element.localName
        self._name = attributeDescriptionValue(getElementAttribute(element,"name"))
        self._attrValue = None
        children = element.childNodes
        for i in range(0,children.length):
            child = children.item(i)
            if child.nodeType == xml.dom.minidom.Node.ELEMENT_NODE:
                if child.localName == "value":
                    if self._attrValue:
                        raise Exception ( "Only one 'value' element allowed: "+elementToXML(child))
                    else:
                        self._attrValue = escapeFilterChars(getElementValue(child))
                else:
                    raise Exception ( "Unknown element: "+elementToXML(child))
            else:
                pass
        if self._attrValue == None:
            raise Exception ( "'"+self._tagName+"' must have a 'value' element: "+elementToXML(element))

        self._ldapFilter = "("+self._name
        if self._type == T_EQUALITY_MATCH:
            self._ldapFilter += "="
        elif self._type == T_GREATER_OR_EQUAL:
            self._ldapFilter += ">="
        elif self._type == T_LESS_OR_EQUAL:
            self._ldapFilter += "<="
        else:
            self._ldapFilter += "~="
        self._ldapFilter += self._attrValue+")"

class MatchingRuleAssertion(Filter):
    """
    Created for extensibleMatch tag

    Parameters:
        element
            Has the form:
                <extensibleMatch dnAttributes="true" name="cn">
                    <value>String</value>
                </extensibleMatch>
            Which, in ldap syntax, translates to:
                (cn:dn:=String)
    """
    def __init__(self,element):
        self._type = typeByTag(element.localName)
        self._tagName = element.localName
        self._name = getElementAttribute(element,"name",False)
        if self._name:
            self._name = attributeDescriptionValue(self._name)
        self._dnAttributes = False

        dnAttributes = getElementAttribute(element,"dnAttributes",False)
        if (dnAttributes != None) & (dnAttributes == "true"):
            self._dnAttributes = True

        self._matchingRule = getElementAttribute(element,"matchingRule",False)

        self._matchValue = None
        children = element.childNodes
        for i in range(0,children.length):
            child = children.item(i)
            if child.nodeType == xml.dom.minidom.Node.ELEMENT_NODE:
                if child.localName == "value":
                    if self._matchValue:
                        raise Exception ( "Only one 'value' element allowed: "+elementToXML(child))
                    else:
                        self._matchValue = escapeFilterChars(getElementValue(child,False))
                else:
                    raise Exception ( "Unknown element: "+elementToXML(child))
            else:
                pass
        if self._matchValue == None:
            raise Exception ( "'"+self._tagName+"' must have a 'value' element: "+elementToXML(element))

        self._ldapFilter = "("
        if self._name:
            self._ldapFilter += self._name
        elif self._matchingRule == None:
            raise Exception ( "Must have a matching rule when no match value given: "+elementToXML(element))
        if self._dnAttributes:
            self._ldapFilter += ":dn"
        if self._matchingRule:
            self._ldapFilter += ":"+self._matchingRule
        self._ldapFilter += ":="+self._matchValue+")"

class SubstringFilter(Filter):
    """
    Created for substrings tag
    """
    def __init__(self,element):
        """
        Create substring filter from dom element

        Parameters:
        element
            Has the form:
                <substrings name="cn">
                    <initial>N</initial>
                    <any>o</any>
                    <any>v</any>
                    <final>l</final>
                </substrings>
            Which, in ldap syntax, translates to:
                (cn=N*o*v*l)
        """
        self._type = typeByTag(element.localName)
        self._tagName = element.localName
        self._ldapFilter = "("+self._tagName+")"
        self._name = attributeDescriptionValue(getElementAttribute(element,"name"))
        self._initial = None
        self._final = None
        self._any = []
        children = element.childNodes
        elementCount = 0
        for i in range(0,children.length):
            child = children.item(i)
            if child.nodeType == xml.dom.minidom.Node.ELEMENT_NODE:
                elementCount += 1
                if child.localName == "initial":
                    if elementCount != 1:
                        raise Exception ( "'initial' tag must be the first child element: "+elementToXML(child))
                    if self._initial:
                        raise Exception ( "Substring filter can only have one 'initial' element: "+elementToXML(child))
                    else:
                        self._initial = escapeFilterChars(getElementValue(child,False))
                elif child.localName == "any":
                    if self._final:
                        raise Exception ( "All 'any' elements must come before the 'final' element: "+elementToXML(child))
                    else:
                        any = escapeFilterChars(child,False)
                        if any:
                            self._any.append(any)
                elif child.localName == "final":
                    if self._final:
                        raise Exception ( "Substring filter can only have one 'final' element: "+elementToXML(child))
                    else:
                        self._final = escapeFilterChars(getElementValue(child,False))
                else:
                    raise Exception ( "Unknown element: "+str(elementToXML(child)))
        self._ldapFilter = "("+self._name+"="
        if self._initial:
            self._ldapFilter += self._initial
        self._ldapFilter += "*"
        for any in self._any:
            self._ldapFilter += any+"*"
        if self._final:
            self._ldapFilter += self._final
        self._ldapFilter += ")"

########################################################################################
## Main
########################################################################################
# process command line arguments
argv = sys.argv[1:]
optlist, args = getopt.getopt(argv, 'hvi:o:l:u:p:f:r:')
for opt, value in optlist:
    if (opt == '-h'):
        printHelp()
        sys.exit(0)
    elif opt == '-v':
        g_Verbose = True
        ldap.set_option(ldap.OPT_DEBUG_LEVEL,1)
    elif opt == '-i':
        g_InputFile = value
    elif opt == '-o':
        g_OutputFile = value
    elif opt == '-l':
        g_LdapUrl = value
    elif opt == '-u':
        g_Username = value
    elif opt == '-p':
        g_Password = value
    elif opt == '-f':
        g_FormattedName = value
    elif opt == '-r':
        g_Request = value

# load secret store, figure out what command line overrides
ss = clstrlibss.load()
if g_LdapUrl == None:
    g_LdapUrl = ss.ldapUrl
if g_Username == None:
    g_Username = ss.adminDn
if g_Password == None:
    g_Password = ss.adminPw

request = None
response = None
batchRequest = None
try:
    # build dom from string passed as arguement, file or stdin
    if g_Request:
        request = xml.dom.minidom.parseString(g_Request)
    elif g_InputFile:
        request = xml.dom.minidom.parse(g_InputFile)
    else:
        request = xml.dom.minidom.parse(sys.stdin)
    debugLog("+#### dsml request ####\n"+request.toxml()+"\n-######################\n")

    # root of document should be a batch request, try and create a BatchRequest
    batchRequest = BatchRequest(request.documentElement)

    # no errors during parsing (wouldn't get here otherwise), bind ldap, service the request, unbind
    if batchRequest._validRequest:
        batchRequest.serviceRequest(g_LdapUrl,g_Username,g_Password)

    # create dom document from result - (verifies correctness, somewhat, for us)
    response = xml.dom.minidom.parseString(batchRequest.response())
except xml.parsers.expat.ExpatError as e:
    response = xml.dom.minidom.parseString("<batchResponse xmlns=\""+g_DSMLv2Namespace+"\"><errorResponse type=\"malformedRequest\"><message>"+str(e)+"</message></errorResponse></batchResponse>")
except Exception as e:
    response = xml.dom.minidom.parseString("<batchResponse xmlns=\""+g_DSMLv2Namespace+"\"><errorResponse type=\"notAttempted\"><message>"+str(e)+"</message></errorResponse></batchResponse>")

# write the response
writeDsmlResponse(response)
