#!/usr/bin/python3

#********************************************************************
#
#  Copyright (c) 2005-2007 Novell, Inc. All Rights Reserved.
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of version 2 of the GNU General Public
#  License as published by the Free Software Foundation.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, contact Novell, Inc.
#
#  To contact Novell about this file by physical or electronic mail,
#  you may find current contact information at www.novell.com
#
#   Author:  Novell, Inc.
#   Workfile:  ncs-resourced.py
#   Contents:  Userspace cluster resource daemon
#
#********************************************************************

import os
import sys
import syslog
import string
import os.path
import ldap
import clstrlib
import clstrlibss
import time
import subprocess
import socket

#############################################################################
#
# Declarations
#

# crmfile = "/tmp/crm"
crmfile = "/proc/ncs/crm"

# OCF attributes
OCF_RA_VERSION_MAJOR = "1"
OCF_RA_VERSION_MINOR = "8"

load = "load"
unload = "unload"
monitor = "monitor"
comatose = "monitor-comatose"
migrate = "monitor-migrate"
reboot = "monitor-reboot"

ldapFailFlag = 0
conf_logrotate_file = "/etc/opt/novell/ncs/ncs-scripts.log.conf"

#############################################################################
#
# Initialization
#

# open syslog
syslog.openlog("ncs-resourced")

# export NCSRESOURCED=1 enables debug
debug = os.getenv("NCSRESOURCED", "disabled")
os.environ["PATH"] ="/sbin" + os.pathsep + os.environ["PATH"]

# initialize LDAP options
ldap.set_option(ldap.OPT_TIMELIMIT, 8)
ldap.set_option(ldap.OPT_NETWORK_TIMEOUT, 16)
certPath = subprocess.Popen(['sed', '-n', 's/TLS_CERT //p', '/etc/openldap/ldap.conf'], stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()[0]
certPath = bytes.decode(certPath).rstrip()
if certPath != '':
    ldap.set_option(ldap.OPT_X_TLS_CACERTFILE, certPath)
    ldap.set_option(ldap.OPT_X_TLS_NEWCTX, 0)
# resource (script) path
resourcePath = "/var/opt/novell/ncs"
ncsRunPath = "/var/run/ncs"
ncsLogPath = "/var/opt/novell/log/ncs"

#############################################################################
#
# Function: resourceToFrom
#
#   Purpose: parse to or from if it is a migrate
#
def resourceToFrom(request):
    try:
        return request.split("=")[4]
    except:
        return None

#############################################################################
#
# Function: resourceCommand
#
#   Purpose: parse resource event from request. ie is this due to a migrate
#
def resourceCommand(request):
    try:
        return request.split("=")[3]
    except:
        return None

#############################################################################
#
# Function: resourceRevision
#
#   Purpose: parse resource revision from request
#
def resourceRevision(request):

    return int(request.split("=")[2])

#############################################################################
#
# Function: resourceName
#
#   Purpose: parse resource name from request
#
def resourceName(request):

    return request.split("=")[1]

#############################################################################
#
# Function: resourceAction
#
#   Purpose: parse resource action from request
#
def resourceAction(request):

    return request.split("=")[0]

#############################################################################
#
# Function: preProcessResourceScript
#
#   Purpose:
#
def preProcessResourceScript(scriptName):

    try:
        try:
            path = os.path.join(ncsRunPath, scriptName)
            f = open(path, "r+")
        except:
            syslog.syslog(syslog.LOG_ERR, " preProcessResourceScript open failed: " + str(sys.exc_info()[0]))
        script = f.read()
        #f.close()
        #f = open(path, "w+")

        changed1, script = clstrlib.translateNetWareScript(script)

        # changed2, script = clstrlib.enableResourceMonitors(script)

        if changed1: # or changed2:
            f.seek(0)
            f.truncate(0)
            f.write(script)

        f.close()

    except Exception as e:
        syslog.syslog(syslog.LOG_WARNING, "Error preprocessing script " + scriptName)



#############################################################################
#
# Function: sendReply
#
#   Purpose: write status to cluster resource manager
#
def sendReply(resName, resAction, resRevision, status):
    reply = resAction + "=" + resName + "=" + str(resRevision) + "=" + str(status)
    f = open(crmfile, "w")
    f.write(reply)
    f.close()

    if (debug != "disabled"):
        syslog.syslog(syslog.LOG_DEBUG, "reply: " + reply)



#############################################################################
#
# Function: resourceRun
#
#   Purpose: run a script file and return it's exit code
#
def resourceRun(resName, resAction, resRevision, timeout=0, resCmd=None, resToFrom=None):
    found, revision = clstrlib.getResourceRevision(resName)
    haveScripts = clstrlib.haveResourceScripts(resName)
    script = resName + "." + resAction
    syslog.syslog(syslog.LOG_INFO, "resourceRun : " + resName)

    if (debug != "disabled"):
        syslog.syslog(syslog.LOG_DEBUG, script + " monitor pid " + str(os.getpid()))

    if (resAction == load):
        if ((resName == "Master_IP_Address_Resource") or
            (found == False) or (resRevision != revision) or (haveScripts == False)):
            # try update resource config
            try:
                # load the secret store
                ss = clstrlibss.load()

                # bind to LDAP
                ld = ldap.initialize(ss.ldapUrl)
                ld.simple_bind_s(ss.adminDn, ss.adminPw)

                # update config for new master
                if (resName == "Master_IP_Address_Resource"):
                    clstrlib.updateClusterConfig(ld, ss.clusterDn)

                # update resName's config
                else:
                    clstrlib.updateResourceConfig(ld, ss.clusterDn, resName)
                    retry = 0
                    found, revision = clstrlib.getResourceRevision(resName)
                    haveScripts = clstrlib.haveResourceScripts(resName)
                    while (retry < 5 and ((found == False) or (resRevision > revision) or (haveScripts == False))):
                        time.sleep( 3 )
                        retry += 1
                        clstrlib.updateResourceConfig(ld, ss.clusterDn, resName)
                        found, revision = clstrlib.getResourceRevision(resName)
                        haveScripts = clstrlib.haveResourceScripts(resName)

                # unbind from LDAP
                ld.unbind_s()

            except:
                syslog.syslog(syslog.LOG_ERR, "LDAP failed: " + str(sys.exc_info()[0]))

        # copy the scripts to the run cache
        os.system("for i in {1..12}; do cp " + os.path.join(resourcePath, resName + ".load") + " " + os.path.join(ncsRunPath, resName + ".load") + "; if [ -s " + os.path.join(ncsRunPath, resName + ".load") + " ]; then break; fi; sleep 1; done")
        os.system("for i in {1..12}; do cp " + os.path.join(resourcePath, resName + ".unload") + " " + os.path.join(ncsRunPath, resName + ".unload") + "; if [ -s " + os.path.join(ncsRunPath, resName + ".unload") + " ]; then break; fi; sleep 1; done")
        os.system("for i in {1..12}; do cp " + os.path.join(resourcePath, resName + ".monitor") + " " + os.path.join(ncsRunPath, resName + ".monitor") + "; if [ -s " + os.path.join(ncsRunPath, resName + ".monitor") + " ]; then break; fi; sleep 1; done")

    # pre-process script
    preProcessResourceScript(script)

    # set OCF attributes
    os.putenv("OCF_RA_VERSION_MAJOR", OCF_RA_VERSION_MAJOR)
    os.putenv("OCF_RA_VERSION_MINOR", OCF_RA_VERSION_MINOR)
    os.putenv("OCF_ROOT", "/usr/lib/ocf" )
    os.putenv("OCF_RESOURCE_INSTANCE", resName)
    os.putenv("OCF_RESOURCE_TYPE", "NCS")

    os.putenv("NCS_COMMAND", str(resCmd))
    os.putenv("NCS_TOFROM", str(resToFrom))

    loadUnloadTimeout = clstrlib.getResourceScriptTimeout( resName, resAction )
    os.putenv("NCS_TIMEOUT", str(loadUnloadTimeout))

    # run script as process
    timeStart = time.time()
    timeTimeout = timeStart + timeout
    logFile = os.path.join(ncsLogPath, script + ".out")
    strTimeStamp = "CRM: " + time.ctime( timeStart )
    os.system( "echo \"" + strTimeStamp + "\" >> " + logFile + "; ( cd " + ncsRunPath + "; dos2unix " + script + " ) &> /dev/null" )
    child = subprocess.Popen( [ "/bin/sh", "-c", os.path.join(ncsRunPath, script) + " >> " + logFile + " 2>&1 </dev/null" ], close_fds=True )

    # read child's status at exit
    if timeout > 0:
        while True:
            status = child.poll()
            if time.time() > timeTimeout:
                status = -999
                break
            if None != status:
                break
            time.sleep(1)
    else:
        status = child.wait()


    # reap defunct children
    try:
        while True:
            cpid, cstatus = os.waitpid(-1, os.WNOHANG)
            if 0 == cpid:
                break;
    except:
        pass

    if (debug != "disabled"):
        syslog.syslog(syslog.LOG_INFO, script + " resource pid " + str(child.pid) + " status " + str(status))

    if status != 0:
        os.system( "grep -m1 -A299 -e \"^" + strTimeStamp + "$\" " + logFile + " | logger -t \"ncs-resourced: " + script + "\"" )

    if (resAction != monitor):
        tmpFile = logFile + ".tmp"
        os.system( "line=`wc -l " + logFile + " 2>/dev/null | grep -oe '^[[:digit:]]\+'`; if [ 1$line -gt 112000 ] ; then tail -6000 " + logFile + " > " + tmpFile + " && mv " + tmpFile + " "  + logFile + " ; fi" )

    return status


#############################################################################
#
# Function: stopMonitor
#
#   Purpose:
#
def stopMonitor(resName):
    try:
        file_name = os.path.join(ncsRunPath, resName + ".pid" )
        pid_file = open( file_name, "r" )
        monitor_pid = pid_file.read()
        pid_file.close()
        os.kill( int( monitor_pid ), 9 )
        os.remove( file_name )
    except:
        pass


#############################################################################
#
# Function: startMonitor
#
#   Purpose:
#
def startMonitor(resName, resRevision):

    status = 0

    cron_jobs = subprocess.Popen(["crontab", "-l"], stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()[0]
    cron_jobs = bytes.decode(cron_jobs)
    #checking if the entry is already existing
    if(cron_jobs.find("novell-ncs-monitor-log") == -1):
        tmpfile = os.path.join(ncsLogPath, "cron_jobs_ncs.tmp")
        monitor_job = "0 0 * * * /usr/sbin/logrotate -f /etc/logrotate.d/novell-ncs-monitor-log > /dev/null 2>&1"
        with open(tmpfile, 'w') as f:
            f.write(cron_jobs + monitor_job + "\n")
            f.close()
        subprocess.call("crontab -u root " + tmpfile + ";rm " + tmpfile, shell=True)

    os.system( "echo '" + str( os.getpid()) + "' > " + os.path.join(ncsRunPath, resName + ".pid" ))
    monTimeout = clstrlib.getResourceMonitorTimeout(resName)
    maxRestarts = clstrlib.getResourceMonitorMaxFailures(resName)
    timePeriod = clstrlib.getResourceMonitorFailurePeriod(resName)
    pollInterval = clstrlib.getResourceMonitorFrequency(resName)
    loadTimeout = clstrlib.getResourceScriptTimeout(resName, load)
    unloadTimeout = clstrlib.getResourceScriptTimeout(resName, unload)
    monitorAction = comatose
    resAction = clstrlib.getResourceMonitorFailover(resName)
    if 1 == resAction:
        monitorAction = migrate
    elif 2 == resAction:
        monitorAction = reboot
    resAction = comatose

    if (debug != "disabled"):
        syslog.syslog(syslog.LOG_DEBUG, "resourceMonitor: " + resName + " maxRestart=" + str(maxRestarts) + " timePeriod=" + str(timePeriod) + " pollInterval=" + str(pollInterval) + " monitorAction=" + str(monitorAction))

    startList = []
    while True:

        while status == 0:
            time.sleep(pollInterval)
            status = resourceRun(resName, monitor, resRevision, monTimeout)

        syslog.syslog(syslog.LOG_INFO, "resourceMonitor: " + resName + ".monitor failed " + str(status))

        timeNow = time.time()
        monitorTime = timeNow - timePeriod
        for timeStarted in startList[:]:
            if timeStarted < monitorTime:
                startList.remove(timeStarted)
        startList.append(timeNow)

        if len(startList) > maxRestarts:
            resAction = monitorAction
            break

        sendReply(resName, "monitor-failure", 0, status)
        syslog.syslog(syslog.LOG_INFO, "resourceMonitor: " + resName + " " + str(len(startList)) + " failures in last " + str(timePeriod) + " seconds")

        status = resourceRun(resName, unload, resRevision, unloadTimeout)
        if status != 0:
            resAction = monitorAction
            syslog.syslog(syslog.LOG_INFO, "resourceMonitor: " + resName + ".unload failed " + str(status))
            break
        sendReply(resName, "monitor-unload", 0, status)

        status = resourceRun(resName, load, resRevision, loadTimeout)
        if status != 0:
            resAction = monitorAction
            syslog.syslog(syslog.LOG_INFO, "resourceMonitor: " + resName + ".load failed " + str(status))
            break
        sendReply(resName, "monitor-load", 0, status)

    syslog.syslog(syslog.LOG_INFO, "resourceMonitor: " + resName + " notifying CRM, action=" + str( resAction ))
    if reboot == resAction:
        delay = 4 + ((clstrlib.getResourceMonitorFrequency( resName ) * (1 + clstrlib.getResourceMonitorMaxFailures( resName ))) >> 6)
        syslog.syslog(syslog.LOG_ERR, "ncs-resourced.py: server rebooting in " + str( delay ) + " seconds." )
        syslog.closelog()
        time.sleep( int( delay ))
        os.system( "echo b > /proc/sysrq-trigger" )

    # todo also give crm the monitorAction
    sendReply(resName, resAction, resRevision, status)

    while True:
        try:
            os.waitpid(-1, os.WNOHANG)
        except:
            break;
        time.sleep( 5 )

    try:
        os.remove( os.path.join(ncsRunPath, resName + ".pid" ))
    except:
        pass

    # exit monitor process
    return 0


#############################################################################
#
# Function: resourceMonitor
#
#   Purpose:
#
def resourceMonitor(request):

    # parse request
    resName = resourceName(request)
    resAction = resourceAction(request)
    resRevision = resourceRevision(request)
    resCmd = resourceCommand(request)
    resToFrom = resourceToFrom(request)

    status = 0

    syslog.syslog(syslog.LOG_DEBUG, "resourceMonitor: " + resName + " " + resAction + " " + str(resRevision))
    if (debug != "disabled"):
        syslog.syslog(syslog.LOG_DEBUG, "resourceMonitor: " + resName + " " + resAction + " " + str(resRevision))

    status = resourceRun(resName, resAction, resRevision, 0, resCmd, resToFrom)
    sendReply(resName, resAction, resRevision, status)
    if status != 0:
        syslog.syslog(syslog.LOG_ERR, "resourceMonitor: " + resName + " " + resAction + " status=" + str(status))

    if (resAction == unload) or (status != 0) or (clstrlib.getResourceMonitorEnabled(resName) == False):
        # exit monitor process
        return 0

    startMonitor(resName, resRevision)

#############################################################################
#
# monitorActions
#
def monitorActions( resName, monitorCmd ):

    status = 0
    revision = 0
    location = ""
    response = -1

    try:
        result = subprocess.Popen(["/sbin/cluster", "status", resName, "--ncs" ], stdout=subprocess.PIPE).communicate()[0].splitlines()
        result  = [ bytes.decode(elem) for elem in result ]
        if result[2].startswith( "Running on" ):
            status = 1
            revision = int( result[3][10:] )
            location = result[2].split( " " )[2]
            resName = result[1].split( " " )[3]
    except:
        pass

    if 0 == status or "" == location or location != socket.gethostname() :
        return "Resource " + resName + " is not running on this node."

    if False == clstrlib.getResourceMonitorEnabled(resName):
        return "Resource monitoring is not enabled for resource " + resName + "."

    if "status" == monitorCmd:
        if os.access( os.path.join(ncsRunPath, resName + ".pid" ), os.F_OK ):
            response = "Resource " + resName + " is being monitored."
        else:
            response = "Resource " + resName + " is not monitored."
    elif "start" == monitorCmd:
        if os.access( os.path.join(ncsRunPath, resName + ".pid" ), os.F_OK ):
            response = "Resource " + resName + " is already monitored."
        else:
            try:
                pid = os.fork()
                if 0 == pid:
                    clstrlib.daemonize()
                    startMonitor( resName, revision )
                    os._exit(0)
                else:
                    if (debug != "disabled"):
                        syslog.syslog(syslog.LOG_DEBUG, "add monitors[\"" + resName + "\"] pid " + str(pid))
                    os.waitpid(pid, 0)
            except OSError as e:
                syslog.syslog(syslog.LOG_ERR, "fork failed: %d (%s)" % (e.errno, e.strerror))

            response = "Started monitoring of resource " + resName + "."
    elif "stop" == monitorCmd:
        if os.access( os.path.join(ncsRunPath, resName + ".pid" ), os.F_OK ):
            stopMonitor( resName )
            response = "Stopped monitoring of resource " + resName + "."
        else:
            response = "Resource " + resName + " is not monitored."
    else:
        response = "Unknow command: " + monitorCmd

    return response

#############################################################################
#
# Main
#
def main(argv=None):
    global ldapFailFlag
    global conf_logrotate_file

    logfile = "/var/opt/novell/log/ncs/*.monitor.out"
    conf_logrotate = logfile + "\n{\n    daily\n    rotate 365\n    olddir /var/opt/novell/log/ncs/archives\n    notifempty\n    compress\n    compresscmd /usr/bin/xz\n    dateext\n    dateformat -%Y-%m-%d_%H:%M:%S\n}\n"

    #writing the configiration to logrotate conf file.
    with open(conf_logrotate_file, 'w') as f:
        f.write(conf_logrotate)
        f.close()

    if argv is None:
        argv = sys.argv

    if( len(argv) > 3 and "monitor" == argv[1].lower()):
        sys.stdout.write(str( monitorActions( argv[2], argv[3] )) + "\n" )
        return 0
    if (len(argv) == 2):
        # override resource script path if supplied
        resourcePath = argv[1]

    # run cluster resource daemon

    clstrlib.daemonize()

    while True:

        # read request from cluster resource manager
        f = open(crmfile, "r")
        request = f.read()
        f.close()

        # handle request
        #
        #    "load=resourceName"
        #    "unload=resourceName"
        #    "exit"
        #    "init"

        if (debug != "disabled"):
            syslog.syslog(syslog.LOG_DEBUG, "request: " + request)

        if (request == "exit"):
            syslog.syslog(syslog.LOG_DEBUG, "recieved exit")
            break
        resName = resourceName(request)
        resAction = resourceAction(request)

        if (resAction == "init" or ldapFailFlag == 1):
            syslog.syslog(syslog.LOG_DEBUG, "recieved init or ldapFailFlag = 1 ...")
            try:
                # load the secret store
                ss = clstrlibss.load()

                # bind to LDAP
                ld = ldap.initialize(ss.ldapUrl)
                ld.simple_bind_s(ss.adminDn, ss.adminPw)

                # init config
                clstrlib.updateClusterConfig(ld, ss.clusterDn)
                # unbind from LDAP
                ld.unbind_s()
                ldapFailFlag = 0

            except:
                syslog.syslog(syslog.LOG_ERR, "LDAP failed: during init ... ")
                ldapFailFlag = 1
            if (resAction == "init"):
                sendReply(resName, resAction, 0, ldapFailFlag)
                continue

        # kill old monitor
        stopMonitor(resName)

        # fork new monitor
        try:
            pid = os.fork()
            if (pid == 0):
                clstrlib.daemonize()
                resourceMonitor(request)
                os._exit(0)
            else:
                if (debug != "disabled"):
                    syslog.syslog(syslog.LOG_DEBUG, "add monitors[\"" + resName + "\"] pid " + str(pid))
                os.waitpid(pid, 0)

        except OSError as e:
            syslog.syslog(syslog.LOG_ERR, "fork failed: %d (%s)" % (e.errno, e.strerror))

    syslog.syslog(syslog.LOG_DEBUG, "exiting")
    syslog.closelog()
    return 0

if __name__ == "__main__":
    sys.exit(main())
