#!/usr/bin/env python3
# © Copyright [2020 - 2021] Micro Focus or one of its affiliates.
#
# The only warranties for products and services of Micro Focus and its affiliates and licensors
# (“Micro Focus”) are as may be set forth in the express warranty statements accompanying such
# products and services. Nothing herein should be construed as constituting an additional
# warranty. Micro Focus shall not be liable for technical or editorial errors or omissions
# contained herein. The information contained herein is subject to change without notice.
#
# Except as specifically indicated otherwise, this document contains confidential information
# and a valid license is required for possession, use or copying. If this work is provided to the
# U.S. Government, consistent with FAR 12.211 and 12.212, Commercial Computer Software,
# Computer Software Documentation, and Technical Data for Commercial Items are licensed
# to the U.S. Government under vendor's standard commercial license.

'''
script to get the information of
	1. the cluster health if the node is the master
	2. the service status if the node is a slave
'''

import json
import subprocess
import os
import socket
import re
import datetime
import fcntl
import sys
import syslog
from collections import OrderedDict
from datetime import datetime

service_name = "novell-ncs"
NCS_ETC = "/etc/opt/novell/ncs"
NCS_VAR="/var/opt/novell/ncs"
NCS_BIN="/opt/novell/ncs/bin"

COMMAND_RUN_ERR = 0x1
SUBPROCESS_ERR = 0x2
FILE_READ_ERR = 0x3
JSON_ERR = 0x4

output_dict = OrderedDict([
			("serviceName", "OES Cluster Services (NCS)"),
			("status", "Healthy"),
			("Type", ""),
			("runningSince", 0),
			("AdditionalInfo", ""),
			("children", None)
			])
output_list = []

cluster_unhealthy = 0


# The dict 'additional_info_key' contains the keys of resource bundle. The below 'keys' (ex: Err_Ncs_ClusterInit)
# will be used by dashboard app to get the localised string.
additional_info_key = { "service_down": "Common_ServiceInactive", "sbd_partition": "Err_Ncs_SbdPartition", "conf_missing" : "Err_Ncs_ConfMissing", "cluster_init_failed": "Err_Ncs_ClusterInit", "node_init_failed": "Err_Ncs_NodeInit", "module_load": "Err_Ncs_Module", "adminfs_mount": "Err_Ncs_AdminfsMount", "daemon_down" : "Common_Daemon", "resource_state1": "Err_Ncs_ResourceState1", "resource_state2": "Err_Ncs_ResourceState2", "resource_state3": "Err_Ncs_ResourceState3", "active_nodes": "Err_Ncs_ActiveNodes", "service_restart": "Common_ResolveService", "res_running": "Info_Ncs_ResourceRunning", "res_offline" : "Info_Ncs_ResourceOffline", "res_comatose" : "Info_Ncs_ResourceComatose", "res_loading" : "Info_Ncs_ResourceLoading", "res_unloading" : "Info_Ncs_ResourceUnloading", "res_unassigned" : "Info_Ncs_ResourceUnassigned", "res_wait" : "Info_Ncs_ResourceWait", "res_sync" : "Info_Ncs_ResourceSync", "res_upgrade": "Info_Ncs_ResourceUprade", "res_startalert": "Info_Ncs_ResourceStartAlert", "res_failoveralert": "Info_Ncs_ResourceFailoverAlert", "res_failbackalert" : "Info_Ncs_ResourceFailbackAlert", "res_unknownalert": "Info_Ncs_ResourceUnknownAlert" }

def return_err(err_str, errno):
	err_dict = OrderedDict()
	err_dict["serviceName"] = "OES Cluster Services (NCS)"
	err_dict["status"] = "Unknown"
#	err_dict["AdditionalInfo"] = err_str
	print(write_json(err_dict))
	sys.exit(errno)

def run_command(cmd_list):
	try:
		proc = subprocess.Popen(cmd_list, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
		out, err = proc.communicate()
		if err:
			err_str = "Error in running the command '" + " ".join(cmd_list) + "'"
			return_err(err_str, COMMAND_RUN_ERR)
		ret = proc.returncode
		return out.decode('utf-8'), ret
	except:
		err_str = "Failed to run the command: " + "'" + " ".join(cmd_list) +  "' using subprocess"
		syslog.syslog(syslog.LOG_ERR, err_str)
		return_err(err_str, SUBPROCESS_ERR)
		sys.exit(SUBPROCESS_ERR)

def systemctl_show_prop(prop):
    output, ret = run_command(["systemctl", "show", "-p", prop, service_name])
    return output.split("=")[1].strip("\n")

def file_read(path):
	try:
		with open(path, "r") as f:
			return f.read()
	except:
		err_str = "Failed to open " + path + " for reading"
		syslog.syslog(syslog.LOG_ERR, err_str)
		return_err(err_str, FILE_OPEN_ERR)

def resource_state_switch(key):
	resource_state = {
		"Running": "res_running",
		"Offline": "res_offline",
		"Loading": "res_loading",
		"Unloading": "res_unloading",
		"Comatose": "res_comatose",
		"Unassigned": "res_unassigned",
		"Quorum Wait" : "res_wait",
		"NDS Sync" : "res_sync",
		"Upgrade" : "res_upgrade",
		"Start": "res_startalert",
		"FailOver" : "res_failoveralert",
		"FailBack" : "res_failbackalert",
		"unknown" : "res_unknownalert"
	}
	return resource_state.get(key)

def get_node_name():
	path = os.path.join(NCS_ETC, "nodename")
	node_name = ""

	if os.path.exists(path):
		node_name = file_read(path)
	if not node_name:
		node_name =  socket.gethostname()

	return node_name

def get_cluster_name():
	cluster_name = ""
	path = os.path.join(NCS_VAR, "cluster.xml")

	if os.path.exists(path) and os.path.getsize(path) > 0:
		grep_proc_output, grep_proc_ret = run_command(['grep', '-A2' ,'name="cn"', path])
		if grep_proc_ret == 0:
			cluster_name = grep_proc_output.split("\n")[2].strip()

	return cluster_name

def module_loaded(module_name):
	"""Checks if module is loaded"""
	lsmod_proc = subprocess.Popen(['lsmod'], stdout=subprocess.PIPE)
	grep_proc = subprocess.Popen(['grep', '-wE', '^' + module_name], stdin=lsmod_proc.stdout, stdout=subprocess.PIPE, stderr = subprocess.PIPE)
	grep_proc.communicate()  # Block until finished
	return grep_proc.returncode == 0

def proc_status(proc):
	pgrep_output, pgrep_ret =  run_command(['pgrep', proc])
	return pgrep_ret == 0


def check_service_health(node_name):
	service_health = []

	cluster_name = get_cluster_name()
	if(cluster_name):
		out, ret = run_command(["sbdutil", "-f", "-n", cluster_name])
	else:
		out, ret = run_command(["sbdutil", "-f"])

	if(ret != 0):
		service_health.append(additional_info_key["sbd_partition"])

	path = os.path.join(NCS_ETC, "clstrlib.conf")
	if not (os.path.exists(path) and os.path.getsize(path) > 0):
		service_health.append("clstrlib.conf," + additional_info_key["conf_missing"])

	path = os.path.join(NCS_VAR, "cluster.xml")
	if not os.path.exists(path):
		service_health.append(additional_info_key["cluster_init_failed"])

	path = os.path.join(NCS_VAR, "nodes.xml")
	if os.path.exists(path) and os.path.getsize(path) > 0:
		grep_output, grep_ret = run_command(["grep", "-i", "dn=\"cn=" + node_name, path])
		if(grep_ret != 0):
			service_health.append(additional_info_key["node_init_failed"])
	else:
		service_health.append("nodes.xml" + "," + additional_info_key["conf_missing"])

	path = os.path.join(NCS_VAR, "gipc.conf")
	if not (os.path.exists(path) and os.path.getsize(path) > 0):
		service_health.append("gipc.conf" + "," + additional_info_key["conf_missing"])

	ncs_modules = ["ncs_timer", "clstrlib", "sbdlib", "vll", "gipc", "sbd", "vipx", "css", "crm", "cmsg", "cma"]
	for module_name in ncs_modules:
		loaded = module_loaded(module_name)
		if not loaded:
			service_health.append(additional_info_key["module_load"] + "," + module_name)
			break

	adminfs_mount_output, adminfs_mount_ret = run_command(["findmnt", "adminfs"])
	if(adminfs_mount_ret != 0):
		service_health.append(additional_info_key["adminfs_mount"])

	ncs_processes = ['adminfsd', 'ncs-configd', 'ncs-resourced', 'ncsdkd', 'ncs-emaild']
	for proc in ncs_processes:
		status = proc_status(proc)
		if not status:
			service_health.append(additional_info_key["daemon_down"] + "," + proc)

	return service_health

def find_match(pattern, string, all_or_one):
	if not all_or_one:
		try:
			output = re.search(pattern, string, flags = re.DOTALL)
		except:
			output = ""
	else:
		try:
			output = re.findall(pattern, string, flags = re.DOTALL)
		except:
			output = ""

	return output

def get_total_nodes():
	total_nodes = []
	path = os.path.join(NCS_VAR, "nodes.xml")

	if os.path.exists(path):
		data = file_read(path)
		pattern = "cn=(.*?),cn="
		total_nodes = re.findall(pattern, data)
	return total_nodes

def get_nodes_info(node_name):
	global cluster_unhealthy, output_list
	nodes_info = []
	cluster_info = dict()
	path = "/admin/Novell/Cluster/NodeState.xml"
	node_type = ""

	if os.path.exists(path):
		xml = file_read(path)
		pattern = "<ncsReply>[\s]*<cluster>(.*?)</cluster>[\s]*<\/ncsReply>"
		xml = find_match(pattern, xml, 0)

		if xml:
			node_type = "slave"  #this node is a part of cluster
			names = ["masterNode", "thisNode", "node"]
			for name in names:
				pattern = "<" + name + ">(.*?)</" + name + ">"
				if name != "node":
					match = find_match(pattern, xml.group(), 0)
					if match:
						cluster_info[name] = match.groups()[0]
					else:
						cluster_info[name] = ""
				else:
					cluster_info["nodes"] = find_match(pattern, xml.group(), 1)

			if cluster_info["masterNode"] and cluster_info["thisNode"]:
				if cluster_info["masterNode"] == cluster_info["thisNode"]:
					node_type = "master"
					total_nodes_list = get_total_nodes()
					total_nodes = len(total_nodes_list)
					active_nodes = len(cluster_info["nodes"])
					if(active_nodes != total_nodes):
						cluster_unhealthy = 1
						fraction_nodes = str(active_nodes) + "/" + str(total_nodes) + "."
						nodes_info.append(additional_info_key["active_nodes"])
						nodes_info.append(fraction_nodes)
					for node in total_nodes_list:
						nodes_status = OrderedDict()
						nodes_status["serviceName"] = node
						nodes_status["Type"] = "node"

						if node == cluster_info["masterNode"]:
							nodes_status["Type"] = "master"

						if node in cluster_info["nodes"]:
							nodes_status["status"] = "Healthy"
						else:
							nodes_status["status"] = "UnHealthy"
						output_list.append(nodes_status)
		else:
			cluster_unhealthy = 1

	return node_type, nodes_info

def get_resource_info():
	global cluster_unhealthy, output_list
	res_info = []

	res_unhealthy = 0
	path = "/admin/Novell/Cluster/ResourceState.xml"

	if os.path.exists(path):
		xml = file_read(path)
		pattern = "<ncsReply>[\s]*<resources>(.*?)</resources>[\s]*</ncsReply>"
		xml = find_match(pattern, xml, 0)
		if xml:
			names = ["name", "state", "alert","location", "upSince"]
			pattern = "<resource>(.*?)</resource>"
			total_res = find_match(pattern, xml.group(), 1)
			for res in total_res:
				res_status = OrderedDict()
				res_status["Type"] = "resource"
				names_dict = dict()
				for name in names:
					pattern = "<" + name + ">(.*?)</" + name + ">"
					match = find_match(pattern, res, 0)
					if match:
						names_dict[name] = match.groups()[0]
					else:
						names_dict[name] = ""
				res_status["serviceName"] = names_dict["name"]
				if(names_dict["state"] == "Running"):
					res_status["status"] = "Healthy"
				elif (names_dict["state"] == "NDS Sync" or names_dict["state"] == "Quorum Wait"):
					res_status["status"] = "Partially Healthy"
				else:
					res_status["status"] = "UnHealthy"

				if(names_dict["state"]):
					res_state = resource_state_switch(names_dict["state"])
					info = additional_info_key["resource_state1"] + "," + additional_info_key[res_state]
					res_status["AdditionalInfo"] = info.split(",")
				elif(names_dict["alert"]):
					res_state = resource_state_switch(names_dict["alert"])
					info = additional_info_key["resource_state1"] + "," + additional_info_key[res_state]
					res_status["AdditionalInfo"] = info.split(",")
				if(names_dict["location"]):
					loc = additional_info_key["resource_state2"] + "," + names_dict["location"]
					res_status["AdditionalInfo"] += loc.split(",")

				if(names_dict["state"] == "Running" and names_dict["upSince"]):
					running_since = names_dict["upSince"].split("=")[-1]
					if running_since:
						res_status["runningSince"] = int(running_since)
				else:
					res_status["runningSince"] = 0
				if names_dict["state"] == "Comatose" or names_dict["alert"]:
						res_unhealthy = 1
				output_list.append(res_status)

		if res_unhealthy == 1:
			res_info.append(additional_info_key["resource_state3"])
			cluster_unhealthy = 1

	return res_info

def write_json(output):
	try:
			return json.dumps(output, separators=(", ", ": "))
	except:
		err_str = "Failed to convert the output to json format"
		syslog.syslog(syslog.LOG_DEBUG, err_str)
		return_err(err_str, JSON_ERR)

def get_ncs_health(node_name):
	global output_list, output_dict

	unit_file_state = systemctl_show_prop("UnitFileState")
	# Don't output anything if the service is disabled
	if(unit_file_state == "disabled"):
		sys.exit(0)

	service_status = systemctl_show_prop("ActiveState")

	if(service_status == "active"):
		timestamp = systemctl_show_prop("ActiveEnterTimestamp")
		if timestamp:
			output_dict["runningSince"] = int(datetime.strptime(timestamp, "%a %Y-%m-%d %H:%M:%S %Z").timestamp())
	else:
		# failed or inactive
		output_dict["status"] = "UnHealthy"
		if(service_status == "inactive"):
			info = "NCS," + additional_info_key["service_down"] + "," + additional_info_key["service_restart"]
			output_dict["AdditionalInfo"] = info.split(",")
			return output_dict
	'''
	control reaches here if the state is active/failed.
	1) if failed try to find the reasons for failure
	2) if active, check if something is going bad
	'''
	service_unhealth_info = check_service_health(node_name)
	if service_unhealth_info:
		output_dict["status"] = "UnHealthy"
	if service_status == "failed":
		output_dict["AdditionalInfo"] =  service_unhealth_info[0].split(",")
		return output_dict
	'''
	state is active, so find if we are the master
	1) if we are the master, get the generic cluster info
	2) if we are the slave, return
	'''
	node_type, nodes_info = get_nodes_info(node_name)
	output_dict["Type"] = node_type
	if node_type == "master":
		res_info = get_resource_info()
		service_unhealth_info += nodes_info + res_info
		output_dict["children"] = output_list
	if cluster_unhealthy:
		output_dict["status"] = "UnHealthy"
		for child in output_dict["children"]:
			if(child["Type"] == "master"):
				child["status"] = "UnHealthy"
				break
	output_dict["AdditionalInfo"] = ",".join(service_unhealth_info).split(",")

	return output_dict

if __name__ == '__main__':
	node_name = get_node_name()
	output = get_ncs_health(node_name)
	print(write_json(output))
