#!/usr/bin/env python3
# © Copyright [2020 - 2021] Micro Focus or one of its affiliates.
#
# The only warranties for products and services of Micro Focus and its affiliates and licensors
# (“Micro Focus”) are as may be set forth in the express warranty statements accompanying such
# products and services. Nothing herein should be construed as constituting an additional
# warranty. Micro Focus shall not be liable for technical or editorial errors or omissions
# contained herein. The information contained herein is subject to change without notice.
#
# Except as specifically indicated otherwise, this document contains confidential information
# and a valid license is required for possession, use or copying. If this work is provided to the
# U.S. Government, consistent with FAR 12.211 and 12.212, Commercial Computer Software,
# Computer Software Documentation, and Technical Data for Commercial Items are licensed
# to the U.S. Government under vendor's standard commercial license.


"""
Python script to fetch NCS data for OES telemetry.
"""

import os
import glob
import subprocess
import time
import json
import re
import sys
from os import path
from collections import OrderedDict
from datetime import datetime

result = OrderedDict([
        ("OESService", "OES Cluster Services (NCS)"),
        ("timestamp", int(time.time())) ])

telemetry_result_loc = "/var/opt/novell/telemetry/data/"
telemetry_config = "/etc/opt/novell/telemetry/config/telemetry.json"

opstrings=["load", "unload"]

#make sure to update both slab_keys and slab_max_val
slab_keys = ["<=15", "<=30", "<=45", "<=60", "<=90", "<=120", "<=150", "<=180", "<=240", "<=300", "<=360", "<=420", "<=480", "<=540", "<=600", ">600" ]
slab_max_val =[ 15,     30,     45,     60,     90,     120,     150,     180,     240,     300,     360,     420,     480,     540,     600, sys.maxsize ]

def map_duration_to_slab_index(secs):

    if len(slab_keys) != len(slab_max_val):
        add_error("keys and corresponding slab maxs are not matching")

    if secs < 0:
        add_error("The duration is negative")
        return len(slab_keys)-1

    #identify the first slot whose slab max value is greater than  duration
    for idx in range(len(slab_max_val)):
        if slab_max_val[idx] >= secs:
            return idx

def add_error( errmsg ):
    """
    This adds all the processing errors occurred into the dictionary
    """
    result["errors"].append(errmsg)

def run_command(cmd):
    ret = 0
    retcode=0
    try:
        ret = subprocess.check_output(cmd, stdin=subprocess.PIPE, stderr=subprocess.PIPE,  shell=True)
        return retcode, ret.decode('ascii')
    except subprocess.CalledProcessError as excobj:
        #add_error("command failed:"+cmd + " ;  errorcode:" + str(excobj.returncode) )
        return excobj.returncode, excobj.output


def add_node_count():
    """
    Function adds the number of nodes in cluster
    """

    node_cnt=0
    cmd = "/sbin/cluster info summary"
    result["node_count"]=0
    retcode, clust_info = run_command( cmd )
    if retcode != 0:
        return
    clust_info_split = clust_info.splitlines(True)
    for line in clust_info_split:
        if "nodeid" in line:
            node_cnt=node_cnt+1
    result["node_count"]=node_cnt


def add_cluster_rsrc_count():
    """
    Function adds the number cluster resources count into dictionary.
    The "Master_IP_Address_Resource" is not included in that count
    """

    cmd ="/sbin/cluster resources "
    result["resource_count"]=0

    retcode, clust_rscs = run_command( cmd )
    if retcode != 0 :
        return
    clust_rscs_split = clust_rscs.splitlines(True)
    rsc_cnt = 0
    start_counting = False
    for line in clust_rscs_split:
        if "Master_IP_Address_Resource" in line:
            start_counting = True
            continue
        #skip counting until master resource info line seen
        if not start_counting :
            continue

        rsc_cnt = rsc_cnt+1

    result["resource_count"]= rsc_cnt

def add_protocol_settings():
    """
    This function add the protocols settings into the dictionary
    """

    cmd = "/sbin/cluster info protocol"
    result["protocol_settings"]= { }
    heading_seen = False
    retcode, clust_info = run_command( cmd )
    if retcode != 0:
        return
    clust_info_split = clust_info.splitlines(True)
    for line in clust_info_split:
        if "Cluster protocol" in line:
            heading_seen = True
            continue
        if not heading_seen:
            continue
        line_split = line.strip().split(",")
        for key_val in line_split:
            pair=key_val.split(":")
            result["protocol_settings"][ pair[0].strip() ]= pair[1].strip()

def add_email_status():
    """
    Function adds email status to the dictionary
    """

    cmd = "/sbin/cluster info notification "
    result["email_notification_enabled"]=False

    retcode, clust_info = run_command( cmd )
    if retcode != 0:
        return
    clust_info_split = clust_info.splitlines(True)
    for line in clust_info_split:
        if "E-mail notification" not in line:
            continue

        # line contains "E-mail notification"
        if "disabled" in line:
            result["email_notification_enabled"]=False
        else:
            result["email_notification_enabled"]=True
        break
    return

def add_sbd_partition_size():
    """
    Function add the sbd is configured or not, size and mirroring information
    """

    result["sbd_partition"] = { }
    size_str ="0MB"
    cmd = "sbdutil -f"
    result["sbd_partition"]["configured"]=False
    result["sbd_partition"]["size"]=size_str
    result["sbd_partition"]["mirrored"]=False
    retcode, sbd_part_path = run_command( cmd )
    # sbd is not configured
    if retcode != 0:
        return

    #sbd is configured
    result["sbd_partition"]["configured"]=True
    cluster_name=path.basename(sbd_part_path).split(".")[0]
    cmd2 = "nlvm list partition "+ path.basename(sbd_part_path)
    retcode, sbd_part_info = run_command( cmd2 )
    if retcode != 0:
        cluster_name=path.basename(sbd_part_path).split(".")[0]
        partition_name = cluster_name + ".msbd0"
        cmd2 = "nlvm list partition "+ partition_name
        retcode, sbd_part_info = run_command( cmd2 )
        if retcode != 0:
            return
        result["sbd_partition"]["mirrored"]=True
    sbd_part_info_split = sbd_part_info.splitlines(True)

    for line in sbd_part_info_split:
        if "Size=" not in line :
            continue
        line = line.strip()
        line_split = line.split(" ")
        for key_val_split in line_split:
            if "Size=" not in key_val_split:
                continue
            val = key_val_split.split("=")
            size_str = val[1].split("(")
            result["sbd_partition"]["size"]=size_str[0]
            break
    return


def add_is_a_mix_node_cluster():
    """
    Function adds 1.cluster is mixed node or not  And
                2.all node's cluster versions
    """

    result["version_info"] = { }
    result["version_info"]["versions"] = []
    result["version_info"]["is_mix_node_cluster"] = False

    summary_sbd_slots = [ ]
    cmd = "/sbin/cluster info summary"
    retcode, clust_info = run_command( cmd )
    if retcode != 0:
        return
    clust_info_split = clust_info.splitlines(True)
    for line in clust_info_split:
        if "nodeid" in line:
            summary_sbd_slots.append( line.split(" ")[-1].strip() )

    cmd = "sbdutil -v"
    retcode, sbd_util_view = run_command( cmd )
    if retcode != 0 :
        return
    sbd_util_view_split = sbd_util_view.split("\n")
    header_seen = False
    for line in sbd_util_view_split:
        if "Signature" in line:
            header_seen = True
            continue
        if not header_seen:
            continue
        if "SBD" not in line:
            continue
        line_split = line.split( )
        signature = line_split[0].strip()
        slot_id = line_split[1]
        if "SBDS" == signature:
            continue
        if slot_id in summary_sbd_slots:
            result["version_info"]["versions"].append( signature )

    first_elem = result["version_info"]["versions"][0]
    for elem in result["version_info"]["versions"]:
        if elem != first_elem :
            result["version_info"]["is_mix_node_cluster"] = True
            return
    return

def add_nds_monitor_status():
    """
    Function add wether NDS monitor enabled
    """

    node_cnt=0
    cmd = "/opt/novell/ncs/install/ncs_install.py -m display  "
    result["nds_monitor_enabled"] = False
    retcode, nds_info = run_command( cmd )
    if retcode != 0:
        return
    nds_info_split = nds_info.splitlines(True)
    for line in nds_info_split:
        if "NDSD monitoring" not in line:
            continue
        if "not enabled" in line:
            result["nds_monitor_enabled"] = False
        else:
            result["nds_monitor_enabled"] = True
        break
    return


def add_resource_monitor_info():
    """
    Functions to count the resources being monitored
    """

    result["resource_monitor_info"] = { }
    result["resource_monitor_info"]["total_resource_count"]= result["resource_count"]
    result["resource_monitor_info"]["online_onthis_node"]= 0
    result["resource_monitor_info"]["monitor_enabled_count"] = 0

    cmd = " /sbin/cluster resources | grep -i `hostname` | grep -v Master_IP_Address_Resource "
    retcode, rsc_onthis_node = run_command( cmd )
    if retcode != 0:
        return
    rsc_onthis_node_split = rsc_onthis_node.splitlines(True)
    result["resource_monitor_info"]["online_onthis_node"]= len(rsc_onthis_node_split)

    monitoring_enabled_count = 0
    for line in rsc_onthis_node_split:
        if "Master_IP_Address_Resource" in line:
            continue
        res_name = line.split()[0]
        cmd = "/sbin/cluster monitor "+ res_name + " status "
        retcode, rsc_monitor_status = run_command( cmd )
        if "is being monitored" in rsc_monitor_status:
            monitoring_enabled_count = monitoring_enabled_count + 1
    result["resource_monitor_info"]["monitor_enabled_count"] = monitoring_enabled_count


def find_duration( crm_line, date_line ):
    """
    Function to deduce duration load or unload episode
    Input : crm_line - complete line containing CRM timestamp with (#Num) signature in the line
            date_line - complete contains the timestamp before exit
    returns the duration in seconds
    """

    try:
        idx = crm_line.find("(")
        crm_date_str = crm_line[:idx]
        crm_date_str = crm_date_str.rstrip(" ")

        idx = date_line.find(" ")
        date_str = date_line[idx+1:].rstrip("\n")

        crm_date_obj = datetime.strptime( crm_date_str, "%a %b %d %H:%M:%S %Y"  )
        date_line_obj = datetime.strptime( date_str, "%a %b %d %H:%M:%S %Z %Y"  )

        duration = date_line_obj - crm_date_obj
        return duration.total_seconds()
    except:
        return -1

def add_in_histogram(crm_line, date_line, resource_name, opstr ):
    """
    Function to update histogram given the two lines containing the begin and end timestamps
    Input : crm_line - complete line containing CRM timestamp with (#Num) signature in the line
            date_line - complete contains the timestamp before exit
            resource_name - This is fake name given for this corresponding resource
            opstr - "load" or "unload"

    The histogram buckets in seconds duration : 15, 30, 45, 60, 90, 120, 150, 180, 240, 300, 360, 400, 420, 480, 540, 600 , >600
    """

    duration_secs = find_duration( crm_line, date_line )
    if duration_secs < 0:
        return
    idx = map_duration_to_slab_index(duration_secs)
    result["histogram"][resource_name][opstr][ slab_keys[idx] ] = result["histogram"][resource_name][opstr][ slab_keys[idx] ]+1


def initialize_histogram_buckets( resource_name ):
    result["histogram"][resource_name]= { }
    for opstr in opstrings:
        result["histogram"][resource_name][opstr] = { }
        for idx in range(len(slab_keys)):
            result["histogram"][resource_name][opstr][ slab_keys[ idx ] ] = 0


def build_time_buckets( resource_output_file, resource_fake_name, opstr ):
    """
    Function - Goes through the dotoutparser output file.
                identifies the lines containing start and end timestamp
                Then updates the histogram for every such matching pair
    Input : resource_output_file - generated using dotoutparser
            resource_fake_name - This is fake name given for this corresponding resource
            opstr - "load" or "unload"
    """
    fp = open(resource_output_file, "r" )
    crm_line = ""
    date_line = ""
    crm_seen = False
    date_seen = False

    while True:
        line = fp.readline()
        #end of file reached
        if not line:
            if crm_seen and date_seen :
                add_in_histogram(crm_line, date_line, resource_fake_name, opstr)
                date_seen = False
            break

        #line doesn't contain - (#Num) and "date"
        if not re.search( r'(#[0-9]*)' , line) and "date" not in line :
            continue

        #line contain (#Num)
        if re.search( r'(#[0-9]*)' , line):
            if crm_seen and date_seen :
                add_in_histogram(crm_line, date_line, resource_fake_name, opstr)
                date_seen = False
            crm_line = line
            crm_seen = True

        #line contains date , need not be exactly the episode last date line
        if "date" in line:
            date_line = fp.readline()
            date_seen = True
    fp.close()
    return

def add_duration_histogram_byname(resource_name, fake_name):
    """
    Function to  generate the dotout paraser output for load and unload operations
    Input : resource_name - The cluster resource name
            fake_name - The fake name choosen for this resource
    """

    initialize_histogram_buckets( fake_name )
    load_out_file_path = "/tmp/"+ resource_name.strip() + opstrings[0] +"_out.txt"
    cmd = "dotoutparser -l -a -vvvv  -i " + resource_name + " >" + load_out_file_path
    run_command( cmd )
    if path.exists( load_out_file_path ):
        build_time_buckets( load_out_file_path, fake_name, opstrings[0])
        cmd = "rm " + load_out_file_path
        run_command(cmd)

    unload_out_file_path = "/tmp/"+ resource_name.strip() + opstrings[1] +"_out.txt"
    cmd = "dotoutparser -u -a -vvvv  -i " + resource_name + " >" + unload_out_file_path
    run_command( cmd )
    if path.exists( unload_out_file_path ):
        build_time_buckets( unload_out_file_path, fake_name, opstrings[1] )
        cmd = "rm " + unload_out_file_path
        run_command(cmd)

def add_duration_histogram():
    """
    Function to the histogram data to the dictionary
    """

    fake_str_name = "Resource"
    cmd ="/sbin/cluster resources "
    retcode, clust_rscs = run_command( cmd )
    if retcode != 0:
        return
    clust_rscs_split = clust_rscs.splitlines(True)

    rsc_cnt = 0
    master_seen = False
    for line in clust_rscs_split:
        #Master resource line seen
        if "Master_IP_Address_Resource" in line:
            master_seen = True
            continue
        #skip until master resource info line seen
        if not master_seen:
            continue

        #fetch the resource name
        rsc_cnt = rsc_cnt+1
        idx = line.find(" ")
        resource_name = line[ :idx]
        fake_name = fake_str_name + str(rsc_cnt)
        add_duration_histogram_byname(resource_name, fake_name )

def telemetry_check():
    try:
        with open(telemetry_config, "r") as file:
            status = json.load(file)
            if status["opt-in"].lower() == "yes":
                return True
    except IOError:
        add_error("Telemetry config file is not accessible")
    return False


def ncs_check():
    cmd = "systemctl is-active novell-ncs"
    retcode, output = run_command( cmd )
    if "active" ==  output.rstrip():
        return True
    return False


def store_result():
    old_files = glob.glob(telemetry_result_loc + 'ncs_*.json')
    for file in old_files:
        try:
            os.remove(file)
        except OSError as e:
            add_error("Unable to delete old data file: " + e.strerror)
    file_name = "ncs_" + str(result["timestamp"]) + ".json"
    file_path = telemetry_result_loc + file_name
    file = open(file_path, "w")
    file.write(json.dumps(result))
    file.close()

def initialize_result_dict():
    """
    Function to take care of any initializations required on the "result" dictionary
    """
    result["histogram"]= { }
    result["errors"]= [ ]

def generate_ncs_data():

    enabled = telemetry_check()
    if not enabled :
        return 0
    ncs_enabled = ncs_check()
    if not ncs_enabled:
        return 0
    initialize_result_dict()
    add_node_count()
    add_cluster_rsrc_count()
    add_protocol_settings()
    add_email_status()
    add_sbd_partition_size()
    add_is_a_mix_node_cluster()
    add_nds_monitor_status()
    add_resource_monitor_info()
    add_duration_histogram()
    store_result()
    return 0

if __name__ == '__main__':
    generate_ncs_data()

