#!/usr/bin/python3

# (C) Copyright 2003, 2018-2020 Micro Focus or one of its affiliates.
# All Rights Reserved.
#
# This work is subject to U.S. and international copyright laws and treaties.
# No part of this work may be used, practiced, performed, copied, distributed,
# revised, modified, translated, abridged, condensed, expanded, collected,
# compiled, linked, recast, transformed or adapted without the prior written
# consent of Novell, Inc. Any use or exploitation of this work without authorization
# could subject the perpetrator to criminal and civil liability.

#import commands
import os
import sys
import getopt
import ldap
import socket #for gethostname
import subprocess
import time
import getpass
import re
import socket

sys.path.append('/opt/novell/ncs/bin')
import clstrlib
import clstrlibss


SUCCESS = 0
EFAILED = 1

OLD_SCHEMA = 0
NEW_SCHEMA = 1

EWRITE_NODENAME = 10
EEXTEND_SCHEMA = 20
ELDAP_BIND = 30
ECREATING_CLUSTER = 40
EADDING_NODE = 50
ECREATING_SBD = 60

SKIP_CONFIG        = 'Skip Configuration'
NEW_CLUSTER        = 'New Cluster'
EXISTING_CLUSTER   = 'Existing Cluster'

START_NOW = 'Now'
START_LATER = 'Later'

LOG_ALL = 0
LOG_DEBUG = 1
LOG_INFO = 2
LOG_ERR = 3

default_ldap_port = 636
dsfw_ldap_port = 1636
max_sbd_devices = 4

log_dir = '/var/opt/novell/install'
log_file = log_dir + '/ncslog'
schema_dir = '/opt/novell/ncs/schema'
config_file = '/etc/sysconfig/novell/ncs'
ncp_config_file = '/etc/sysconfig/novell/ncpsrvr'
edir_config_file = '/etc/opt/novell/eDirectory/conf/nds.conf'
dsfw_config_file = '/etc/sysconfig/novell/xad'
clstrlib_config = '/etc/opt/novell/ncs/clstrlib.conf'
ncsldif_file = schema_dir + '/ncs.ldif'
ncssch_file = schema_dir + '/ncs.sch'
ncpldif_file = schema_dir + '/ncpserver.ldif'
preldif_file = schema_dir + '/ncpserver.preldif'
lvm_conf_file = '/etc/lvm/lvm.conf'
lvm_local_conf_file = '/etc/lvm/lvmlocal.conf'

log = None #file object
doItAll = True
extendSchema = False
createClusterObj = False
addNodeToCluster = False
deleteNodeFromCluster = False
createNodeNameFile = False
createLibConfFile = False
createSBD = False
startRequested = False
setProxyUser = False
isNCPEnabled = False
isDsfwEnabled = False
addFirewallRichRule = False
addLvmSanlock = False

ndsd_monitor_config = ''
ndsd_monitor_frequency = 60
ndsd_monitor_timeout = 20

ldap.set_option(ldap.OPT_NETWORK_TIMEOUT,8)
ldap.set_option(ldap.OPT_TIMEOUT,16)
certPath = subprocess.Popen(['sed', '-n', 's/TLS_CERT //p', '/etc/openldap/ldap.conf'], stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()[0]
certPath = bytes.decode(certPath).rstrip()
if certPath != '':
    ldap.set_option(ldap.OPT_X_TLS_CACERTFILE, certPath)
    ldap.set_option(ldap.OPT_X_TLS_NEWCTX, 0)
#############################################################################
#
# Function: log_append
#
# Returns: Nothing.
#
# Comments:
#   write an entry to the log file
#
def log_append(level, text):
    global log

    if level >= get_log_level():
        if not log:
            os.system('mkdir -p '+log_dir)
            log = open(log_file, 'a+')
        if log:
            log.write( time.strftime('%x %X: ' + text + '\n'))


#############################################################################
#
# Function: nds2ldap
# Converts a nds context to an ldap context (replaces . with ,)
# Param: s The nds type context
# Returns: the ldap equivelent of s
#
def nds2ldap(s):
    return s.replace('.',',')


#############################################################################
#
# Function: ldap2nds
# Converts a ldap context to a nds context (replaces , with .)
# Param: s The ldap type context
# Returns: the nds equivelent of s
#
def ldap2nds(s):
    return s.replace(',','.')


#############################################################################
#
# Function: parse_command_line
#
# Returns: Nothing.
#
# Comments:
#   parse the command line and set global variables.
#
def parse_command_line():
    global config_file
    global dsfw_config_file
    global doItAll
    global extendSchema
    global createClusterObj
    global addNodeToCluster
    global deleteNodeFromCluster
    global createNodeNameFile
    global createLibConfFile
    global createSBD
    global startRequested
    global setProxyUser
    global ndsd_monitor_config
    global ndsd_monitor_frequency
    global ndsd_monitor_timeout
    global addFirewallRichRule
    global addLvmSanlock

    log_append(LOG_INFO, '\n-----------------------------------------------')
    log_append(LOG_INFO, '%s' %sys.argv)
    sysconfig_file = config_file
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'f: d: i: t: m: a b c e l n r s u p g', ['configfile', 'dsfwconfigfile' ])
    except getopt.GetoptError:
        sys.exit(3)
    for o, a in opts:
        if o in ('-f', '--configfile'):
            if not os.path.isfile(a):
                log_append(LOG_ERR, 'Could not find config file %s' % a)
                return EFAILED
            config_file = a
        elif o in ('-d', '--dsfwconfigfile'):
            if not os.path.isfile(a):
                log_append(LOG_ERR, 'Could not find dsfw config file %s' % a)
                return EFAILED
            dsfw_config_file = a
        elif o in ('-a'):
            # print "Add node"
            doItAll = False
            addNodeToCluster = True
        elif o in ('-c'):
            # print "Create cluster object in edir"
            doItAll = False
            createClusterObj = True
        elif o in ('-e'):
            # print "Extend schema"
            doItAll = False
            extendSchema = True
        elif o in ('-l'):
            # print "Create clstrlib.conf"
            doItAll = False
            createLibConfFile = True
        elif o in ('-n'):
            # print "Create node name file"
            doItAll = False
            createNodeNameFile = True
        elif o in ('-b'):
            # print "Create SBD"
            doItAll = False
            createSBD = True
        elif o in ('-s'):
            # print "Start NCS"
            doItAll = False
            startRequested = True
        elif o in ('-u'):
            setProxyUser = True
        elif o in ('-r'):
            try:
                os.system( "sed -i 's/^[[:space:]]*SERVICE_CONFIGURED\\>.*$/SERVICE_CONFIGURED=\"no\"/' " + sysconfig_file + " &>/dev/null" )
            except:
                pass
            doItAll = False
            deleteNodeFromCluster = True
        elif o in ('-m'):
            doItAll = False
            ndsd_monitor_config = a
        elif o in ('-i'):
            doItAll = False
            ndsd_monitor_frequency = a
        elif o  in ('-t'):
            doItAll = False
            ndsd_monitor_timeout = a
        elif o in ('-p'):
            doItAll = False
            addFirewallRichRule = True
        elif o in ('-g'):
            doItAll = False
            addLvmSanlock = True

    if doItAll:
        extendSchema = True
        createClusterObj = True
        addNodeToCluster = True
        createNodeNameFile = True
        createLibConfFile = True
        createSBD = True
        startRequested = True
        addFirewallRichRule = True
        addLvmSanlock = True

    return SUCCESS


#############################################################################
#
# Function: get_log_level
#
# Returns: path to the log file
#
def get_log_level():
    return LOG_ALL


#############################################################################
#
# Function: run
#
# Executes a command with os.system.  You have control over the
# log level, whether the command line gets logged (to avoid logging
# passwords), and the log level.
# note that you can't get the output for processing.
#
# Parameters: cmd The command to be executed
#             logCmd: 1 the command is logged, 0 it is not (passwords)
#             The log level.
#
# Returns: return code of os.system(cmd)
#
def run(cmd, logCmd = True, logLvl = LOG_INFO):
    global log_file
    global log_dir
    rcfile = log_dir + '/ncsrc'

    theCmd  = cmd

    if logCmd:
        log_append(logLvl, 'run: ' + theCmd)

    if get_log_level() >= logLvl:
        theCmd = '( %s ; echo $? > %s ) >> %s 2>&1' % (cmd, rcfile, log_file)
    else:
        theCmd = '( %s ; echo $? > %s ) > /dev/null 2>&1' % (cmd, rcfile)

    os.system(theCmd)

    # open and read the return value from the tmp file.
    try:
        file = open(rcfile, 'r')
    except:
        log_append(LOG_ERR, 'Error in run getting return value, defaulting to 0')
        return 0

    rc = file.readline()
    file.close()
    os.system('rm -rf %s' % rcfile)
    return int(rc)

#############################################################################
#
# Function: run_with_output
#
# Return:
#   (rc, rstr) This is a tuple that contains the return code
#   (rc) and the output from the command (rstr).
#
# Comments:
#   Same as run but also returns the output.
#
def run_with_output(cmd, logCmd = True, logLvl = LOG_INFO):
    if logCmd:
        log_append(logLvl, 'run_with_output: ' + cmd)

    try:
        rstr = subprocess.Popen( [ "/bin/sh", "-c", cmd ], stderr=subprocess.PIPE, stdout=subprocess.PIPE, close_fds=True ).communicate()[0]
    except:
        rstr = ""

    if get_log_level() >= logLvl:
        log_append(logLvl, rstr)

    return (rc, rstr)


#############################################################################
#
# Method:   read_config_file
#
# Returns: SUCESS or EFAILED
#
# Comments:
#   This function reads in the configuration file and stores the
#   information in the config dictionary.
#
#   The file format is: <key>=<value>
#   If the first character in a line is a '#', it is considered a
#   comment.
#
def read_config_file(file_name, conf):
    # open the configuration file
    log_append(LOG_DEBUG, 'Opening the configuration file (%s)' % file_name)
    try:
        file = open(file_name, 'r')
    except IOError as message:
        log_append(LOG_INFO, 'cannot open the conf file (%s) for reading' % file_name)
        log_append(LOG_INFO, '  system message: %s' % message)
        return EFAILED

    # read the configuration file
    log_append(LOG_DEBUG, 'Reading the configuration file (%s)' % file_name)
    try:
        lines = file.readlines()
    except IOError as message:
        log_append(LOG_INFO, 'cannot read the conf file (%s) for reading' % file_name)
        log_append(LOG_INFO, '  system message: %s' % message)

    # place the contents of the file in the configuration database (dictionary)
    for line in lines:
        if line[0] == '#':
            continue
        if line.count('=') != 0:
            option = line.split('=', 1)       # parse out key & value
            option[0] = option[0].strip()     # remove any leading & trailing whitespace
            option[1] = option[1].strip()     # ditto
            option[1] = option[1].strip('\"') # now kill any " chars
            conf[option[0]] = option[1]       # set the config dictionary value
    file.close()
    return SUCCESS


#############################################################################
#
# Method: get_config_file
#
# Returns: config value form the config database
#
# Comments:
#
def get_config_value(conf, valueName):
    try:
        r = conf[valueName]
    except:
        r = ''
    return r


#############################################################################
#
# Method: get_primary_ip
#
# Returns: string with this machines primary ip address
#
# Comments:
#
def get_primary_ip():
    return socket.gethostbyname(socket.gethostname())


#############################################################################
#
# Method: get_dsfw_settings
#
# Returns: nothing
#
# Comments:
#
def get_dsfw_settings():

    global isDsfwEnabled

    dsfwconf= get_config_value(dsfw_config, 'SERVICE_CONFIGURED')
    log_append(LOG_DEBUG, 'get_dsfw_settings() - dsfwconf is: %s.' %dsfwconf)
    if dsfwconf == 'yes' or dsfwconf == 'YES' or dsfwconf == 'Yes':
        isDsfwEnabled = True
    else:
        isDsfwEnabled = False


    log_append(LOG_DEBUG, 'get_dsfw_settings() - isDsfwEnabled is: %s.' %isDsfwEnabled)

    return SUCCESS


#############################################################################
#
# Method: get_ncp_settings
#
# Returns: nothing
#
# Comments:
#
def get_ncp_settings():

    global isNCPEnabled

    ncpconf= get_config_value(ncp_config, 'SERVICE_CONFIGURED')
    log_append(LOG_DEBUG, 'get_ncp_settings() - ncpconf is: %s.' %ncpconf)
    if ncpconf == 'yes' or ncpconf == 'YES' or ncpconf == 'Yes':
        isNCPEnabled = True
    else:
        isNCPEnabled = False

    log_append(LOG_DEBUG, 'get_ncp_settings() - isNCPEnabled is: %s.' %isNCPEnabled)

    return SUCCESS


#############################################################################
#
# Method: get_ldap_settings
#
# Returns: Noting.
#
# Comments:
#
def get_ldap_settings():
    global ldapIp
    global ldapPort
    global ldapServers

    ldapServers = get_config_value(config, 'CONFIG_NCS_LDAP_INFO')
    log_append(LOG_DEBUG, 'get_ldap_settings() - ldap server list is: %s.' %ldapServers)
    if ldapServers == '':
        log_append(LOG_DEBUG, 'get_ldap_settings() - empty list')
        ldapIp = get_primary_ip()
        ldapPort = default_ldap_port
        ldapServers = "ldaps://%s:%s" % (ldapIp, ldapPort)

    else:
        log_append(LOG_DEBUG, 'get_ldap_settings() - parse list')
        i = ldapServers.find(',')
        if i < 8:
            i = len(ldapServers)
        firstServer = ldapServers[8:i]
        part = firstServer.split(':')
        ldapIp = part[0]
        if len(part) < 2 or part[1] == '':
            ldapPort = default_ldap_port
        else:
            ldapPort = part[1]

    log_append(LOG_DEBUG, 'get_ldap_settings() - ldap ip is: %s.' %ldapIp)
    log_append(LOG_DEBUG, 'get_ldap_settings() - ldap port is: %s.' %ldapPort)
    log_append(LOG_DEBUG, 'get_ldap_settings() - ldap server list is: %s.' %ldapServers)

    return SUCCESS

#############################################################################
#
# Method: get_servername
#
# Returns: string with this machines hostname
#
# Comments:
#
def get_servername():
    return socket.gethostname();

#############################################################################
#
# Method: get_serverDN
#
# Returns: string with this machines DN (e.g. cn=servername,ou=here,o=novell)
#
# Comments:
#
def get_serverDn():

    global isNCPEnabled

    if isNCPEnabled:
        srvname = get_config_value(edir_config, 'n4u.nds.server-name')
        srvcontext = get_config_value(edir_config, 'n4u.nds.server-context')
        srvcontext = nds2ldap(srvcontext)

        srvdn = "cn=%s,%s" % (srvname, srvcontext)

    else:
        srvdn = ''

    log_append(LOG_DEBUG, 'get_serverDN() - returns : %s.' %srvdn)

    return srvdn;

#############################################################################
#
# Method: get_config_type
#
# Returns: If this should be a new cluster or and excisting cluster
#
# Comments: Does an ldap search to see if the cluster object already excists.
#
def get_config_type():
    ldapList = ldapServers.split( ',' )
    for server in ldapList:
        parts = server.split( '//', 1 )[1].split(':', 1)
        ldapIp = parts[0];
        if 2 > len( parts ) or '' == parts[1]:
            ldapPort = '636'
        else:
            ldapPort = parts[1]

        cmd = "/usr/bin/ldapsearch -H ldaps://%s:%s -D %s -w %s -x -s base -LLL -b %s objectClass" % (ldapIp, ldapPort, adminDn, '*****', clusterDn)
        log_append(LOG_INFO, cmd)
        cmd = "/usr/bin/ldapsearch -H ldaps://%s:%s -D %s -w '%s' -x -s base -LLL -b %s objectClass" % (ldapIp, ldapPort, adminDn, adminPw, clusterDn)
        rc = run(cmd, False)
        if rc != SUCCESS and isDsfwEnabled:
            cmd = "/usr/bin/ldapsearch -H ldaps://%s:%s -D %s -w %s -x -s base -LLL -b %s objectClass" % (ldapIp, dsfw_ldap_port, adminDn, '*****', clusterDn)
            log_append(LOG_INFO, cmd)
            cmd = "/usr/bin/ldapsearch -H ldaps://%s:%s -D %s -w '%s' -x -s base -LLL -b %s objectClass" % (ldapIp, dsfw_ldap_port, adminDn, adminPw, clusterDn)
            rc = run(cmd, False)

        if rc == SUCCESS:
            return EXISTING_CLUSTER

    return NEW_CLUSTER

#############################################################################
#
# Method: get_schema_status
#
# Returns: If the schema should be updated
#
# Comments: Does an ldap search to see if the schema object has a certain value
#
def get_schema_status():
    testString = 'nCSCRMMonitorAction'

    ldapList = ldapServers.split( ',' )
    for server in ldapList:
        parts = server.split( '//', 1 )[1].split(':', 1)
        ldapIp = parts[0];
        if 2 > len( parts ) or '' == parts[1]:
            ldapPort = '636'
        else:
            ldapPort = parts[1]

        cmd = "/usr/bin/ldapsearch -H ldaps://%s:%s -D %s -w %s -x -s base -LLL -b cn=schema | grep %s" % (ldapIp, ldapPort, adminDn, '*****', testString)
        log_append(LOG_INFO, cmd)
        cmd = "/usr/bin/ldapsearch -H ldaps://%s:%s -D '%s' -w '%s' -x -s base -LLL -b cn=schema | grep %s" % (ldapIp, ldapPort, adminDn, adminPw, testString)
        rc = run(cmd, False)
        if rc == SUCCESS:
            return NEW_SCHEMA

    return OLD_SCHEMA

#############################################################################
#
# Method: verify_settings
#
# Returns: Nothing.
#
# Comments: Verifies the install settings and put them in global variables
#           Currently Only verifies LDAP settings.

def verify_settings():

   log_append(LOG_INFO, 'Verifying LDAP Binding to Server ...')
   ld = ldap.initialize(ldapServers)

   if ld:
       try:
          ld.simple_bind_s(adminDn, adminPw)
       except ldap.LDAPError as e:
          log_append(LOG_ERR, 'NCS ldap bind failed. Invalid LDAP server or User Credential')
          log_append(LOG_ERR, str(e))
          return ELDAP_BIND

   return SUCCESS



#############################################################################
#
# Method: get_settings
#
# Returns: Nothing.
#
# Comments: Pull the install settings and put them in global variables
#
def get_settings():
    global configType
    global ldapIp
    global ldapPort
    global ldapServers
    global adminDn
    global adminPw
    global serverName
    global serverDn
    global serverIp
    global clusterDn
    global clusterIp
    global clusterContainer
    global clusterName
    global isShared
    global start
    global sbdDevices
    global sbdSize
    global isDsfwEnabled
    global proxyUsername
    global proxyPassword

    sbdDevices = []

    clusterDn = nds2ldap(get_config_value(config, 'CONFIG_NCS_CLUSTER_DN'))

    get_ldap_settings()

    adminDn = nds2ldap(get_config_value(config, 'CONFIG_NCS_ADMIN_DN'))
    adminPw = get_config_value(config, 'CONFIG_NCS_ADMIN_PASSWORD')
    # If we are being called by YaST, then we need to get the password from the secure tmp file.
    if adminPw == '':
        tmpdir = ''
        pFile = ''
        log_append(LOG_ERR, 'ncs_install:Entering with an empty password.')
        try:
            tmpdir = open('/var/opt/novell/install/tmpdir', 'r')
            content = tmpdir.readlines()
            tmpPath = content[0].strip()
            tmpPath = tmpPath + '/ncsfile'
            #log_append(LOG_ERR, 'ncs_install:pword file path is: %s.'%tmpPath)
            tmpdir.close()

            pFile = open(tmpPath, 'r')
            content = pFile.readlines()
            adminPw = content[0].strip()
            pFile.close()
        except (IOError, IndexError):
            if tmpdir:
                tmpdir.close()
            if pFile:
                pFile.close()
            if adminPw == '':
                log_append(LOG_ERR, 'NCS failed to get password.')

    if '' == adminDn:
        print ("Please enter the DN for the admin (cn=admin,o=novell): ")
        adminDn = nds2ldap( sys.stdin.readline().strip())

        if '' == adminDn:
            adminDn = "cn=admin,o=novell"
        adminPw = getpass.unix_getpass( "please enter the password for " + adminDn + ": " )

    if setProxyUser:
        proxyUsername = nds2ldap(get_config_value(config, 'CONFIG_NCS_MGT_USERNAME'))
        if '' == proxyUsername:
            print ("Please enter the DN for NCS management user (cn=ncs_proxy_user,o=novell): ")
            proxyUsername  = nds2ldap(sys.stdin.readline().strip())

        proxyPassword = get_config_value(config, 'CONFIG_NCS_MGT_PASSWORD')
        if '' == proxyPassword:
            tmpdir = ''
            pFile = ''
            log_append(LOG_ERR, 'ncs_install:Entering with an empty proxy password.')
            try:
                tmpdir = open('/var/opt/novell/install/tmpdir', 'r')
                content = tmpdir.readlines()
                tmpPath = content[0].strip()
                tmpPath = tmpPath + '/ncsfile'
                tmpdir.close()

                pFile = open(tmpPath, 'r')
                content = pFile.readlines()
                if 1 < len( content ):
                    proxyPassword = content[1].strip()
                pFile.close()
            except (IOError, IndexError):
                if tmpdir:
                    tmpdir.close()
                if pFile:
                    pFile.close()
                if '' == proxyPassword:
                    log_append(LOG_ERR, 'NCS failed to get proxy password.')

        if '' == proxyPassword:
            proxyPassword = getpass.unix_getpass( "please enter the password for " + proxyUsername + ": " )

    serverIp = get_config_value(config, 'CONFIG_NCS_SERVER_IP')
    if serverIp == '':
        serverIp = get_primary_ip()

    serverName = get_config_value(config, 'CONFIG_NCS_SERVER_NAME')
    if serverName == '':
        serverName = get_servername();

    serverDn = get_serverDn();

    configType = get_config_value(config, 'CONFIG_NCS_CONFIG_TYPE')
    if configType == '':
        configType = get_config_type()

    if configType == NEW_CLUSTER:
        clusterIp = get_config_value(config, 'CONFIG_NCS_CLUSTER_IP')
        (rc, clusterName, clusterContainer) = extract_cn(clusterDn)

        sbdDevice = get_config_value(config, 'CONFIG_NCS_SBD_DEV')
        if sbdDevice:
            sbdDevices.append(sbdDevice)

        for i in range(max_sbd_devices):
            sbdDevice = get_config_value(config, 'CONFIG_NCS_SBD_DEV' + str(i))
            if sbdDevice:
                sbdDevices.append(sbdDevice)

        if len(sbdDevices):
            isShared = True
            sbdSize = get_config_value(config, 'CONFIG_NCS_SBD_SIZE')
        else:
            isShared = False

    start = get_config_value(config, 'CONFIG_NCS_START')

    return SUCCESS


#############################################################################
#
# Method: build_ncp_attr_list
#
# Returns: SUCCESS or error code
#
# Comments:
#
def build_ncp_attr_list(ld):
    res = ld.search_s('cn=schema', ldap.SCOPE_BASE)
    attrs = res[0][1]["objectClasses"]
    for a in attrs:
        if isinstance(a, bytes):
            findstr = b'2.16.840.1.113719.1.1.6.1.10'
        else:
            findstr = '2.16.840.1.113719.1.1.6.1.10'
        if a.find(findstr) >= 0:
            attr = a

    try:
        preldif = open(preldif_file, 'r')
        content = preldif.readlines()
        preldif.close()

        ncsattrs = content[0][2:len(content[0])-1]

        ncpldif = open(ncpldif_file, 'w')
        ncpldif.writelines(content)
        ncpldif.seek(-1, 2)

        i = attr.find("MAY ( ") + 6
        ncpldif.write(attr[0:i] + ncsattrs + attr[i:] + "\n")

        ncpldif.close()
    except (IOError, IndexError):
        return EEXTEND_SCHEMA

    return SUCCESS

#############################################################################
#
# Method: extend_schema_ldif
#
# Returns: SUCCESS or an error code
#
# Comments: accepts an ldif file
#
def extend_schema_ldif(ldapIp, ldapPort, adminDn, adminPw, schemaFile, schemaFileFormat):
    exe = '/opt/novell/eDirectory/bin/ice'

    log_append(LOG_INFO, "Extending LDIF schema for NCS ...")

    ldapList = ldapServers.split( ',' )
    for server in ldapList:
        parts = server.split( '//', 1 )[1].split(':', 1)
        ldapIp = parts[0];
        if 2 > len( parts ) or '' == parts[1]:
            ldapPort = '636'
        else:
            ldapPort = parts[1]

        cmd = 'LDAPTLS_REQCERT="Never" ADM_DEST_PASSWD="%s" %s -S %s -f %s -D LDAP -s %s -p %s -d "%s"' % ('*****', exe, schemaFileFormat, schemaFile, ldapIp, ldapPort, adminDn)
        log_append(LOG_INFO, cmd)

        cmd = 'LDAPTLS_REQCERT="Never" ADM_DEST_PASSWD="%s" %s -S %s -f %s -D LDAP -s %s -p %s -d "%s"' % (adminPw, exe, schemaFileFormat, schemaFile, ldapIp, ldapPort, adminDn)
        rc = run(cmd, False)
        if SUCCESS == rc:
            return rc

    return rc

#############################################################################
#
# Method: apply_settings
#
# Returns: Nothing.
#
# Comments:
#
def apply_settings():
    global  clusterDn

    log_append(LOG_INFO, '*** Starting NCS Post-Install ***')

    log_append(LOG_DEBUG, 'config type is: %s.' %configType)
    log_append(LOG_DEBUG, 'ldap ip is: %s.' %ldapIp)
    log_append(LOG_DEBUG, 'ldap port is: %s.' %ldapPort)
    log_append(LOG_DEBUG, 'ldap server list is: %s.' %ldapServers)
    log_append(LOG_DEBUG, 'Admin dn is: %s.' %adminDn)
    log_append(LOG_DEBUG, 'serverName is: %s.' %serverName)
    log_append(LOG_DEBUG, 'serverDn is: %s.' %serverDn)
    log_append(LOG_DEBUG, 'server IP is: %s.' %serverIp)
    log_append(LOG_DEBUG, 'cluster dn is: %s.' %clusterDn)

    log_append(LOG_DEBUG, 'Extend schema flag is: %s.' %extendSchema)
    log_append(LOG_DEBUG, 'Create cluster object flag is: %s.' %createClusterObj)
    log_append(LOG_DEBUG, 'Add node flag is: %s.' %addNodeToCluster)
    log_append(LOG_DEBUG, 'Delete node flag is: %s.' %deleteNodeFromCluster)
    log_append(LOG_DEBUG, 'Create node name file flag is: %s.' %createNodeNameFile)
    log_append(LOG_DEBUG, 'Create cluster lib config file flag is: %s.' %createLibConfFile)
    log_append(LOG_DEBUG, 'Create SBD flag is: %s.' %createSBD)
    log_append(LOG_DEBUG, 'Add rich rule flag is: %s.' %addFirewallRichRule)
    log_append(LOG_DEBUG, 'Enable sanlock for ncs lvm2 volumes: %s.' %addLvmSanlock)
    log_append(LOG_DEBUG, 'Start requested flag is: %s.' %startRequested)

    log_append(LOG_DEBUG, 'DSFW Enabled is: %s.' %isDsfwEnabled)
    log_append(LOG_DEBUG, 'Set Proxy User: %s.' %setProxyUser)

    log_append(LOG_DEBUG, 'ndsd_monitor_config is: %s.' %ndsd_monitor_config)
    log_append(LOG_DEBUG, 'ndsd_monitor_frequency is: %s.' %ndsd_monitor_frequency)
    log_append(LOG_DEBUG, 'ndsd_monitor_timeout is: %s.' %ndsd_monitor_timeout)

    if setProxyUser:
        log_append(LOG_DEBUG, 'proxy username is: %s.' % proxyUsername)

    if configType == NEW_CLUSTER:
        log_append(LOG_DEBUG, 'clusterIp is: %s.' %clusterIp)
        for i in range(len(sbdDevices)):
            log_append(LOG_DEBUG, 'sbdDevice[%d] is: %s.' % (i, sbdDevices[i]))

    ## Create clstrlib.conf
    if os.path.isfile( clstrlib_config ) and 0 < os.path.getsize( clstrlib_config ):
        c = clstrlibss.load()
    else:
        c = clstrlibss.data()
    c.ldapUrl = ldapServers
    if setProxyUser:
        c.adminDn = proxyUsername
        c.adminPw = proxyPassword
    else:
        c.adminDn = adminDn
        c.adminPw = adminPw
    if '' != clusterDn:
        c.clusterDn = clusterDn

    if createLibConfFile:
        clstrlibss.save(c)

    if createNodeNameFile:
        ## Create nodename file
        cmd = 'echo -n ' + serverName + ' > /etc/opt/novell/ncs/nodename'
        rc = run(cmd)
        if rc != 0:
            log_append(LOG_ERR, 'NCS install failed writing node name.')
            return EWRITE_NODENAME

    if createClusterObj or addNodeToCluster or deleteNodeFromCluster or extendSchema or setProxyUser:
        ## Bind to LDAP
        log_append(LOG_INFO, 'Binding to LDAP server ...')
        ld = ldap.initialize(c.ldapUrl)
        if ld:
            tryCount = 1
            maxTry = 5
            while True:
                try:
                    ld.simple_bind_s(adminDn, adminPw)
                except ldap.LDAPError as e:
                    if tryCount >= maxTry:
                        log_append(LOG_ERR, 'NCS ldap bind failed')
                        log_append(LOG_ERR, str(e))
                        return ELDAP_BIND
                    else:
                        log_append(LOG_ERR, 'NCS ldap bind failed. Will try again')
                        tryCount = tryCount + 1
                        os.system('sleep 5')
                        continue
                break

            if extendSchema:
                ## Extend the LDAP schema
                build_ncp_attr_list(ld)

                rc = extend_schema_ldif(ldapIp, ldapPort, adminDn, adminPw, ncsldif_file, "LDIF")
                if rc != SUCCESS:
                    log_append(LOG_INFO, 'NCS install failed extending schema using ldif for ncs. rc=' + str(rc))
                    schemaStatus = get_schema_status()
                    if schemaStatus == OLD_SCHEMA:
                        log_append(LOG_ERR, 'NCS install failed extending schema using ldif.')
                        ld.unbind()
                        return EEXTEND_SCHEMA
                    else:
                        log_append(LOG_INFO, 'Schema (ncs) up to date.')

                rc = extend_schema_ldif(ldapIp, ldapPort, adminDn, adminPw, ncssch_file, "SCH")
                if rc != SUCCESS:
                    log_append(LOG_INFO, 'NCS install failed extending schema using sch for ncs. rc=' + str(rc))
                    schemaStatus = get_schema_status()
                    if schemaStatus == OLD_SCHEMA:
                        log_append(LOG_ERR, 'NCS install failed extending schema using sch.')
                        ld.unbind()
                        return EEXTEND_SCHEMA
                    else:
                        log_append(LOG_INFO, 'Schema (ncs) up to date.')

                rc = extend_schema_ldif(ldapIp, ldapPort, adminDn, adminPw, ncpldif_file, "LDIF")
                if rc != SUCCESS:
                    log_append(LOG_INFO, 'NCS install failed extending schema for ncpServer. rc=' + str(rc))
#                    return EEXTEND_SCHEMA
                else:
                    log_append(LOG_INFO, 'Schema (ncpServer) up to date.')

            if createClusterObj:
                ## Create cluster object
                tryCount = 1
                if configType == NEW_CLUSTER:
                    log_append(LOG_INFO, 'Creating cluster object ...')
                    while True:
                        try:
                            clstrlib.createCluster(ld,
                                                   clusterContainer,
                                                   clusterName,
                                                   clusterIp,
                                                   isShared)
                        except Exception as e:
                            log_append(LOG_ERR, str(e))
                            traceback.print_exc()
                            if tryCount >= maxTry:
                                log_append(LOG_ERR, 'NCS create cluster failed. Aborting Install')
                                ld.unbind()
                                return ECREATING_CLUSTER
                            else:
                                log_append(LOG_ERR, 'NCS create cluster failed. Retrying')
                                tryCount = tryCount + 1
                                os.system('sleep 5')
                                continue
                        break

            if addNodeToCluster:
                ## Create node object
                try:
                    result = ld.search_s(c.clusterDn, ldap.SCOPE_BASE)[0][0]
                    if result and result != clusterDn:
                        c.clusterDn = result
                        clstrlibss.save(c)
                        clusterDn = result
                except Exception as e:
                    c.clusterDn = clusterDn

                tryCount = 1
                log_append(LOG_INFO, 'Adding server to cluster ...')
                while True:
                    try:
                        clstrlib.addServerToCluster(ld,
                                                    c.clusterDn,
                                                    serverDn,
                                                    serverName,
                                                    serverIp)
                    except Exception as e:
                        if tryCount >= maxTry:
                            log_append(LOG_ERR, 'NCS adding node failed: %s' % e)
                            traceback.print_exc()
                            ld.unbind()
                            return EADDING_NODE
                        else:
                            log_append(LOG_ERR, 'NCS adding node failed. Will try again')
                            tryCount = tryCount + 1
                            os.system('sleep 5')
                            continue
                    break

            if setProxyUser:
                if 0 != clstrlib.addUserToNCSGroup( ld, c.clusterDn, proxyUsername ):
                    log_append(LOG_ERR, 'NCS install failed to assign proxy user: %s.' % proxyUsername )
                    ld.unbind()
                    return EFAILED

            if deleteNodeFromCluster:
                try:
                    log_append(LOG_INFO, "Deleting server '%s' from cluster '%s' ..." % (serverDn, c.clusterDn ))
                    clstrlib.deleteServerFromCluster( ld, c.clusterDn, serverDn )
                    log_append(LOG_INFO, "Server '%s' deleted from cluster '%s'." % (serverDn, c.clusterDn ))
                    try:
                        disable_proc = subprocess.Popen(["systemctl", "disable", "novell-ncs"], stdout = subprocess.PIPE, stderr = subprocess.PIPE)
                        err = disable_proc.communicate()[1]
                        if disable_proc.returncode != 0:
                            log_append("Error when disabling the service: %s." % err)
                    except:
                        pass
                except Exception as e:
                    log_append(LOG_ERR, "NCS install failed to delete server '%s' from cluster: %s." % (serverDn, str(e)))
                    ld.unbind()
                    return EFAILED

            ld.unbind()

    if createSBD:
        ## Create SBD partition
        if configType == NEW_CLUSTER and isShared:
            log_append(LOG_INFO, 'Creating SBD partition ...')
            cmd = '/opt/novell/ncs/bin/sbdutil -c -n ' + clusterName
            for i in range(min(len(sbdDevices), max_sbd_devices - 1)):
                cmd += ' -d ' + sbdDevices[i]
            if sbdSize:
                cmd += ' -s ' + str( sbdSize )
                log_append(LOG_INFO, 'SBD partition size ' +  str( sbdSize) + ' MB' )
            rc = run(cmd)
            if rc != 0:
                log_append(LOG_ERR, 'NCS install failed to create SBD partition.')
                return ECREATING_SBD

    if startRequested:
        ## Start NCS
        if start == START_NOW:
            rc = run('systemctl start novell-ncs.service')

    log_append(LOG_INFO, '*** Ending NCS Post-Install ***')

    return SUCCESS


#############################################################################
#
# Function: modify_ndsd_monitor_settings
#
# Returns: SUCCESS or EFAILED
#
# Comments:
#
def modify_ndsd_monitor_settings():
    if os.path.isfile( clstrlib_config ) and 0 < os.path.getsize( clstrlib_config ):
        ss = clstrlibss.load()
    else:
        ss = clstrlibss.data()

    if 'disable' == ndsd_monitor_config:
        ss.ndsdMonitor = 'R 0 0'
        clstrlibss.save( ss )
        os.system( "/sbin/killproc NCS_NDSD_monitord &>/dev/null")
    elif 'reboot-node' == ndsd_monitor_config:
        ss.ndsdMonitor = 'B ' + ndsd_monitor_frequency + ' ' + ndsd_monitor_timeout
        clstrlibss.save( ss )
        os.system( "/sbin/killproc NCS_NDSD_monitord &>/dev/null")
        os.system( "/opt/novell/ncs/bin/NCS_NDSD_monitord " + ss.ndsdMonitor )
    elif 'hard-reboot' == ndsd_monitor_config:
        ss.ndsdMonitor = 'E ' + ndsd_monitor_frequency + ' ' + ndsd_monitor_timeout
        clstrlibss.save( ss )
        os.system( "/sbin/killproc NCS_NDSD_monitord &>/dev/null")
        os.system( "/opt/novell/ncs/bin/NCS_NDSD_monitord " + ss.ndsdMonitor )
    elif 'restart-ndsd' == ndsd_monitor_config:
        ss.ndsdMonitor = 'R ' + ndsd_monitor_frequency + ' ' + ndsd_monitor_timeout
        clstrlibss.save( ss )
        os.system( "/sbin/killproc NCS_NDSD_monitord &>/dev/null")
        os.system( "/opt/novell/ncs/bin/NCS_NDSD_monitord " + ss.ndsdMonitor )
    elif 'get-monitor-action' == ndsd_monitor_config:
        ss = clstrlibss.load()
        if hasattr( ss, 'ndsdMonitor' ):
            sys.stdout.write("NDSD Monitor value: " + ss.ndsdMonitor + "\n")
            return SUCCESS
        else:
            sys.stdout.write( "NDSD monitoring is not enabled (1).\n" )
            return SUCCESS
    elif 'display' != ndsd_monitor_config:
        sys.stdout.write( "Usage: " + sys.argv[0] + " -m display|disable|restart-ndsd|reboot-node|hard-reboot|get-monitor-action|help [-i interval_value] [-t timeout_value]\n" )
        sys.stdout.write( "  -m display: Display the current NDSD monitoring settings\n" )
        sys.stdout.write( "  -m disable: Disable NDSD monitoring\n" )
        sys.stdout.write( "  -m restart-ndsd: Enable NDSD monitoring and set remedy action to restarting NDSD\n" )
        sys.stdout.write( "  -m reboot-node: Enable NDSD monitoring and set remedy action to rebooting node\n" )
        sys.stdout.write( "  -m hard-reboot: Enable NDSD monitoring and set remedy action to rebooting node hard\n" )
        sys.stdout.write( "  -m get-monitor-action: Display the NDSD monitor action\n" )
        sys.stdout.write( "  -m help: Print this help page\n" )
        sys.stdout.write( "  -i interval_value: Specify in seconds how frequently the monitor process checks NDSD\n" )
        sys.stdout.write( "  -t timeout_value: Specify in seconds how long the monitor process waits for the checking operation\n" )

    ss = clstrlibss.load()
    if not hasattr( ss, 'ndsdMonitor' ):
        sys.stdout.write( "NDSD monitoring is not enabled (1).\n" )
    else:
        res = re.match( '^([A-Z]) ([0-9]+) ([0-9]+)$', ss.ndsdMonitor, 0 )
        if not res:
            sys.stdout.write( "NDSD monitoring is not enabled (2).\n" )
        else:
            if '0' == res.group(2):
                sys.stdout.write( "NDSD monitoring is not enabled (3).\n" )
            else:
                sys.stdout.write( "NDSD monitoring is enabled.\n" )
                sys.stdout.write( "interval = " + res.group(2) + " seconds, timeout = " + res.group(3) + " seconds ")
                if 'B' == res.group(1):
                    sys.stdout.write( " and remedy action is rebooting node\n" )
                elif 'E' == res.group(1):
                    sys.stdout.write( " and remedy action is rebooting node hard.\n" )
                else:
                    sys.stdout.write( " and remedy action is restarting ndsd.\n" )

    return SUCCESS

#############################################################################
#
# Function: extract_cn
#
# Returns: CN from FDN
#
# Comments:
#
def extract_cn(fdn):
    i = fdn.find(',')
    if i < 4:
        return (EFAILED, '', '')
    #skip over 'cn='
    cn = fdn[3:i]
    cont = fdn[i+1:]
    return (SUCCESS, cn, cont)

def open_firewall_port():
    isfirewallrunning = os.system('firewall-cmd --state &>/dev/null')

    if(isfirewallrunning != 0):
       cmd = 'firewall-offline-cmd --add-rich-rule=\'rule family=\"ipv4\" protocol value=\"224\" accept\' &>/dev/null'
       ret = os.system(cmd)
    else:
       cmd = 'firewall-cmd --permanent --add-rich-rule=\'rule family=\"ipv4\" protocol value=\"224\" accept\' &>/dev/null'
       ret = os.system(cmd)
       os.system('firewall-cmd --reload &>/dev/null')

    if(ret != 0):
       sys.stdout.write(f'Failed to add NCS protocol 224 to firewall, cmd is {cmd}, error code = {ret}')
       return ret

    return SUCCESS

def enable_SanlockForLvm():

    #Enable Lvm2-lockd, Sanlock services
    isEnabled = os.system('systemctl is-enabled lvmlockd.service &>/dev/null')
    if(isEnabled != 0):
        cmd = 'systemctl enable lvmlockd.service &>/dev/null'
        ret = os.system(cmd)
    isEnabled = os.system('systemctl is-enabled sanlock.service &>/dev/null')
    if(isEnabled != 0):
        cmd = 'systemctl enable sanlock.service &>/dev/null'
        ret = os.system(cmd)
    isEnabled = os.system('systemctl is-enabled wdmd.service &>/dev/null')
    if(isEnabled != 0):
        cmd = 'systemctl enable wdmd.service &>/dev/null'
        ret = os.system(cmd)
    #update the config files
    cmd =  "sed -i '/^#/!s/use_lvmlockd = 0/use_lvmlockd = 1/' " + lvm_conf_file + " &>/dev/null"
    ret = os.system(cmd)
    cmd =  "sed -i 's/use_lvmlockd = [0-9]*/use_lvmlockd = 1/g' " + lvm_conf_file + " &>/dev/null"
    ret = os.system(cmd)
    cmd =  "sed -i 's/locking_type = [0-9]*/locking_type = 1/g' " + lvm_conf_file + " &>/dev/null"
    ret = os.system(cmd)
    cmd =  "sed -i '/zero_metadata/{/^#/! s/^/# /}' " + lvm_conf_file + " &>/dev/null"
    ret = os.system(cmd)
    #set local/hostid in lvmlocal.conf range<1 to 2000>
    hostname = socket.gethostname()
    IPAddr = socket.gethostbyname(hostname)

    #grep nodeid and increment
    results = subprocess.Popen( ["/bin/sh", "-c", "grep '^nodeid " + IPAddr + "' /var/opt/novell/ncs/gipc.conf | awk '{print $3}'" ], close_fds=True, stdout=subprocess.PIPE ).communicate()[0];
    results = bytes.decode(results)
    if(results):
        results = str.strip(results)
        results =  int(results) + int('1')
        cmd = " sed -i  's/# host_id = 0/host_id = " + str(results) + "/g' " +  lvm_local_conf_file + " &>/dev/null"
        ret = os.system(cmd)
    return SUCCESS

#############################################################################
#
# Function: main
#
# Returns: Nothing.
#
# Comments:
#
def main():
    global config_file
    global config
    global configType

    global dsfw_config_file
    global dsfw_config
    global dsfw_enabled

    global ncp_config_file
    global ncp_config
    global ncp_enabled

    global edir_config_file
    global edir_config

    config = dict()
    dsfw_config = dict()
    ncp_config = dict()
    edir_config = dict()

    rc = parse_command_line()
    if rc != SUCCESS:
        sys.exit(rc)

    if addLvmSanlock:
        rc = enable_SanlockForLvm()
        if rc != SUCCESS:
            sys.exit(rc)

    if doItAll == False and addLvmSanlock == True:
        sys.exit(rc)

    if addFirewallRichRule:
        rc = open_firewall_port()
        if rc != SUCCESS:
            sys.exit(rc)

    if doItAll == False and addFirewallRichRule == True:
        sys.exit(rc)

    if '' != ndsd_monitor_config:
        sys.exit( modify_ndsd_monitor_settings())

    rc = read_config_file(dsfw_config_file, dsfw_config)
    if rc == SUCCESS:
        get_dsfw_settings()
    else:
        isDsfwEnabled = False

    rc = read_config_file(ncp_config_file, ncp_config)
    if rc == SUCCESS:
        get_ncp_settings()
        read_config_file(edir_config_file, edir_config)
    else:
        isNCPEnabled = False

    rc = read_config_file(config_file, config)
    if rc == SUCCESS:
        get_settings()
    else:
        sys.exit(rc)

    rc = verify_settings()
    if rc != SUCCESS:
        log_append(LOG_DEBUG, 'ncs_install: Config verification failed (exiting..): ')
        sys.exit(rc)

    rc = 0
    if configType != '' and configType != SKIP_CONFIG:
        rc = apply_settings()

    sys.exit(rc)


if __name__ == "__main__":
    sys.exit(main())

